@@ -1,6 +1,6 @@
 /*
   Copyright (C) 2007, 2008, 2010 Jeroen Frijters
-  Copyright (C) 2009 Volker Berlin (i-net software)
+  Copyright (C) 2009 - 2012 Volker Berlin (i-net software)
   Copyright (C) 2010 Karsten Heinrich (i-net software)
 
   This software is provided 'as-is', without any express or implied
@@ -55,18 +55,23 @@ namespace IKVM.NativeCode.sun.awt.shell
 		private const uint IMAGE_BITMAP = 0;
 		private const uint IMAGE_ICON = 1;
 
+        private const int HINST_COMMCTRL = -1;
+        private const int IDB_VIEW_SMALL_COLOR = 4;
+
+        private const uint WM_USER = 0x0400;
+        private const uint TB_GETIMAGELIST = WM_USER + 49;
+        private const uint TB_LOADIMAGES = WM_USER + 50;
+
+        private const uint ILD_TRANSPARENT = 0x00000001;
+
 		private static readonly IntPtr hmodShell32;
-		private static readonly IntPtr hmodComctl32;
 		private static readonly bool isXP;
-		private static readonly bool isVista;
 
 		[System.Security.SecuritySafeCritical]
 		static Win32ShellFolder2()
 		{
 			hmodShell32 = LoadLibrary("shell32.dll");
-			hmodComctl32 = LoadLibrary("comctl32.dll");
 			isXP = Environment.OSVersion.Version >= new Version(5, 1);
-			isVista = Environment.OSVersion.Version.Major >= 6;
 		}
 
 		[System.Security.SecurityCritical]
@@ -135,7 +140,39 @@ namespace IKVM.NativeCode.sun.awt.shell
 		[DllImport("user32.dll")]
 		static extern bool GetIconInfo(IntPtr hIcon, out ICONINFO piconinfo);
 
-		[StructLayout(LayoutKind.Sequential)]
+        [DllImport("user32.dll", SetLastError = true)]
+        static extern IntPtr CreateWindowEx(
+           uint dwExStyle,
+           string lpClassName,
+           string lpWindowName,
+           uint dwStyle,
+           int x,
+           int y,
+           int nWidth,
+           int nHeight,
+           IntPtr hWndParent,
+           IntPtr hMenu,
+           IntPtr hInstance,
+           IntPtr lpParam);
+
+        [DllImport("user32.dll", CharSet = CharSet.Unicode, SetLastError = true)]
+        [return: MarshalAs(UnmanagedType.Bool)]
+        static extern bool DestroyWindow(IntPtr hwnd);
+
+        [DllImport("user32.dll")]
+        static extern IntPtr SendMessage(IntPtr hWnd, UInt32 Msg, Int32 wParam, Int32 lParam);
+
+        [DllImport("comctl32.dll", SetLastError = true)]
+        public static extern IntPtr ImageList_GetIcon(IntPtr himl, int i, uint flags);
+
+        [DllImport("comctl32")]
+        [return: MarshalAs(UnmanagedType.Bool)]
+        static extern bool ImageList_Destroy(IntPtr himl);
+
+        [DllImport("user32")]
+        public static extern int DestroyIcon(IntPtr hIcon);
+
+        [StructLayout(LayoutKind.Sequential)]
 		private struct BITMAPINFO
 		{
 			internal uint biSize;
@@ -301,60 +338,37 @@ namespace IKVM.NativeCode.sun.awt.shell
             }
         }
 
-		[System.Security.SecuritySafeCritical]
-		public static Bitmap getFileChooserBitmap()
-		{
-			if (hmodShell32 != IntPtr.Zero)
-			{
-				// Code copied from ShellFolder2.cpp Java_sun_awt_shell_Win32ShellFolder2_getFileChooserBitmapBits
-				// Get a handle to an icon.
-				SafeGdiObjectHandle hBitmap = null;
-				try
-				{
-					hBitmap = isVista ?
-						 LoadImage(hmodShell32, "IDB_TB_SH_DEF_16", IMAGE_BITMAP, 0, 0, 0) :
-						 LoadImage(hmodShell32, (IntPtr)216, IMAGE_BITMAP, 0, 0, 0);
-					if (hBitmap == null && hmodComctl32 != IntPtr.Zero)
-					{
-						hBitmap = LoadImage(hmodComctl32, (IntPtr)124, IMAGE_BITMAP, 0, 0, 0);
-					}
-					if (hBitmap != null)
-					{
-                        BITMAPINFO bmi = new BITMAPINFO();
-                        GetObject(hBitmap, Marshal.SizeOf(bmi), ref bmi);
-                        int width = bmi.biWidth;
-                        int height = bmi.biHeight;
-                        bmi.biSize = 40;
-                        bmi.biHeight = -bmi.biHeight;
-                        bmi.biPlanes = 1;
-                        bmi.biBitCount = 32;
-                        bmi.biCompression = 0;
-                        if (width == 0 || height == 0)
+        // Code copied from Java_sun_awt_shell_Win32ShellFolder2_getStandardViewButton0
+        [System.Security.SecuritySafeCritical]
+        public static Bitmap getStandardViewButton0(int iconIndex)
+        {
+            Bitmap result = null;
+            using (new ThemingActivationContext())
+            {
+                // Create a toolbar
+                IntPtr hWndToolbar = CreateWindowEx(0, "ToolbarWindow32", null, 0, 0, 0, 0, 0, IntPtr.Zero, IntPtr.Zero, IntPtr.Zero, IntPtr.Zero);
+                if (hWndToolbar != IntPtr.Zero)
+                {
+                    SendMessage(hWndToolbar, TB_LOADIMAGES, IDB_VIEW_SMALL_COLOR, HINST_COMMCTRL);
+
+                    IntPtr hImageList = SendMessage(hWndToolbar, TB_GETIMAGELIST, 0, 0);
+                    if (hImageList != IntPtr.Zero)
+                    {
+                        IntPtr hIcon = ImageList_GetIcon(hImageList, iconIndex, ILD_TRANSPARENT);
+                        if (hIcon != IntPtr.Zero)
                         {
-                            return null;
+                            Icon icon = Icon.FromHandle(hIcon);
+                            result = icon.ToBitmap();
+                            icon.Dispose();
+                            DestroyIcon(hIcon);
                         }
-                        using (SafeDeviceContextHandle dc = SafeDeviceContextHandle.Get())
-                        {
-                            int[] data = new int[width * height];
-                            GetDIBits(dc, hBitmap, (uint)0, (uint)height, data, ref bmi, 0);
-                            Bitmap bitmap = new Bitmap(data.Length / 16, 16, PixelFormat.Format32bppArgb);
-                            BitmapData bitmapData = bitmap.LockBits(new Rectangle(0, 0, bitmap.Width, bitmap.Height), ImageLockMode.WriteOnly, PixelFormat.Format32bppArgb);
-                            Marshal.Copy(data, 0, bitmapData.Scan0, data.Length);
-                            bitmap.UnlockBits(bitmapData);
-                            return bitmap;
-                        }
-					}
-				}
-				finally
-				{
-					if (hBitmap != null)
-					{
-						hBitmap.Close();
-					}
-				}
-			}
-			return null;
-		}
+                        ImageList_Destroy(hImageList);
+                    }
+                    DestroyWindow(hWndToolbar);
+                }
+            }
+            return result;
+        }
 
         /// <summary>
         /// Retrieves an icon from the shell32.dell
@@ -362,13 +376,14 @@ namespace IKVM.NativeCode.sun.awt.shell
         /// <param name="iconID">the index of the icon</param>
         /// <returns>The icon or null, if there is no icon at the given index</returns>
 		[System.Security.SecuritySafeCritical]
-		public static Bitmap getShell32IconResourceAsBitmap(int iconID)
+        public static Bitmap getShell32IconResourceAsBitmap(int iconID, bool getLargeIcon)
 		{
 			if (hmodShell32 == IntPtr.Zero)
 			{
 				return null;
 			}
-            using (SafeGdiObjectHandle hicon = LoadImage(hmodShell32, (IntPtr)iconID, IMAGE_ICON, 16, 16, 0))
+            int size = getLargeIcon ? 32 : 16;
+            using (SafeGdiObjectHandle hicon = LoadImage(hmodShell32, (IntPtr)iconID, IMAGE_ICON, size, size, 0))
 			{
 				if (hicon != null)
 				{
@@ -381,6 +396,7 @@ namespace IKVM.NativeCode.sun.awt.shell
         /// <summary>
         /// Returns the pIDL of the desktop itself
         /// </summary>
+        [System.Security.SecurityCritical]
         public static IntPtr initDesktopPIDL()
         {
             IntPtr pidl = new IntPtr();
@@ -394,6 +410,7 @@ namespace IKVM.NativeCode.sun.awt.shell
         /// <summary>
         /// Returns an IShellFolder for the desktop
         /// </summary>
+        [System.Security.SecurityCritical]
         public static Object initDesktopFolder()
         {
             ShellApi.IShellFolder rootShell = null;
@@ -410,6 +427,7 @@ namespace IKVM.NativeCode.sun.awt.shell
         /// <param name="desktopIShellFolder">The IShellFolder instance of the Desktop</param>
         /// <param name="csidl">The CSIDL of the special folder</param>
         /// <returns>the desktop relative pIDL of a special folder</returns>
+        [System.Security.SecurityCritical]
         public static IntPtr initSpecialPIDL(Object desktopIShellFolder, int csidl)
         {
             IntPtr result = new IntPtr();        
@@ -423,6 +441,7 @@ namespace IKVM.NativeCode.sun.awt.shell
         /// <param name="desktopIShellFolder">The IShellFolder instance of the Desktop</param>
         /// <param name="pidl">The desktop relative pIDL of the special folder</param>
         /// <returns>The IShellFolder for a special folder</returns>
+        [System.Security.SecurityCritical]
         public static Object initSpecialFolder(Object desktopIShellFolder, IntPtr pidl)
         {
             try
@@ -547,7 +566,7 @@ namespace IKVM.NativeCode.sun.awt.shell
             {
                 return;
             }
-            Marshal.Release(pIDL);
+            Marshal.FreeCoTaskMem(pIDL);
         }
 
 		/// <summary>
@@ -589,7 +608,8 @@ namespace IKVM.NativeCode.sun.awt.shell
             return (int)atts[0];
         }
 
-        public static String getFileSystemPath(int csidl)
+        [System.Security.SecuritySafeCritical]
+        public static String getFileSystemPath0(int csidl)
         {
             IntPtr pIDL = new IntPtr();
             int hRes = ShellApi.SHGetSpecialFolderLocation(IntPtr.Zero, (ShellApi.CSIDL)csidl, ref pIDL);
@@ -609,6 +629,8 @@ namespace IKVM.NativeCode.sun.awt.shell
                 return null;
             }
         }
+
+        [System.Security.SecurityCritical]
         public static Object getEnumObjects(Object pIShellFolder, Boolean isDesktop, Boolean includeHiddenFiles)
         {
             if (pIShellFolder == null)
@@ -632,6 +654,7 @@ namespace IKVM.NativeCode.sun.awt.shell
         /// </summary>
         /// <param name="pEnumObjects">The IEnumIDList to get the next element of</param>
         /// <returns>a pIDL or IntPtr.Zero in case the end of the enum is reached</returns>
+        [System.Security.SecurityCritical]
         public static IntPtr getNextChild(Object pEnumObjects)
         {
             if (pEnumObjects == null)
@@ -690,6 +713,7 @@ namespace IKVM.NativeCode.sun.awt.shell
         /// <param name="pIShellFolder">The IShellFolder to get the chilf of</param>
         /// <param name="name">The display name of the child</param>
         /// <returns>the relative pIDL of the child or IntPrt.Zero in case there is no such child</returns>
+        [System.Security.SecurityCritical]
         public static IntPtr parseDisplayName0(Object pIShellFolder, String name)
         {
             if (pIShellFolder == null)
@@ -756,7 +780,8 @@ namespace IKVM.NativeCode.sun.awt.shell
             return 0;
         }
 
-		public static IntPtr getIcon(String absolutePath, Boolean getLargeIcon) 
+        [System.Security.SecurityCritical]
+        public static IntPtr getIcon(String absolutePath, Boolean getLargeIcon) 
         {
             ShellApi.SHFILEINFO shinfo = new ShellApi.SHFILEINFO();
             if (ShellApi.SHGetFileInfo(absolutePath, 0, out shinfo, (uint)Marshal.SizeOf(shinfo), ShellApi.SHGFI.SHGFI_ICON | (getLargeIcon ? ShellApi.SHGFI.SHGFI_LARGEICON : ShellApi.SHGFI.SHGFI_SMALLICON)) == IntPtr.Zero)
