@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997, 2007, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 1997, 2013, Oracle and/or its affiliates. All rights reserved.
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
  * This code is free software; you can redistribute it and/or modify it
@@ -73,7 +73,7 @@ import java.awt.image.WritableRaster;
  * bounds of the virtual device.  The following code sample
  * calculates the bounds of the virtual device.
  *
- * <pre>
+ * <pre>{@code
  *      Rectangle virtualBounds = new Rectangle();
  *      GraphicsEnvironment ge = GraphicsEnvironment.
  *              getLocalGraphicsEnvironment();
@@ -87,7 +87,7 @@ import java.awt.image.WritableRaster;
  *              virtualBounds =
  *                  virtualBounds.union(gc[i].getBounds());
  *          }
- *      } </pre>
+ *      } }</pre>
  *
  * @see Window
  * @see Frame
@@ -142,7 +142,13 @@ public abstract class GraphicsConfiguration {
      * @return a <code>BufferedImage</code> whose data layout and color
      * model is compatible with this <code>GraphicsConfiguration</code>.
      */
-    public abstract BufferedImage createCompatibleImage(int width, int height);
+    public BufferedImage createCompatibleImage(int width, int height) {
+        ColorModel model = getColorModel();
+        WritableRaster raster =
+            model.createCompatibleWritableRaster(width, height);
+        return new BufferedImage(model, raster,
+                                 model.isAlphaPremultiplied(), null);
+    }
 
     /**
      * Returns a <code>BufferedImage</code> that supports the specified
@@ -296,7 +302,7 @@ public abstract class GraphicsConfiguration {
     public VolatileImage createCompatibleVolatileImage(int width, int height,
         ImageCapabilities caps, int transparency) throws AWTException
     {
-        VolatileImage vi = 
+        VolatileImage vi =
             new Container().createVolatileImage(width, height, caps);
         if (caps != null && caps.isAccelerated() &&
             !vi.getCapabilities().isAccelerated())
@@ -426,4 +432,21 @@ public abstract class GraphicsConfiguration {
         }
         return defaultImageCaps;
     }
+
+    /**
+     * Returns whether this {@code GraphicsConfiguration} supports
+     * the {@link GraphicsDevice.WindowTranslucency#PERPIXEL_TRANSLUCENT
+     * PERPIXEL_TRANSLUCENT} kind of translucency.
+     *
+     * @return whether the given GraphicsConfiguration supports
+     *         the translucency effects.
+     *
+     * @see Window#setBackground(Color)
+     *
+     * @since 1.7
+     */
+    public boolean isTranslucencyCapable() {
+        // Overridden in subclasses
+        return false;
     }
+}
