@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2003, 2012, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
  * This code is free software; you can redistribute it and/or modify it
@@ -31,7 +31,8 @@ import cli.System.IO.FileShare;
 import cli.System.IO.FileStream;
 import cli.System.IO.SeekOrigin;
 import cli.System.Runtime.InteropServices.DllImportAttribute;
-import java.util.concurrent.atomic.AtomicInteger;
+import java.util.ArrayList;
+import java.util.List;
 
 /**
  * Instances of the file descriptor class serve as an opaque handle
@@ -57,11 +58,15 @@ public final class FileDescriptor {
      *   JRuby uses reflection to get at the handle (on Windows) and fd (on non-Windows)
      *   fields to convert it into a native handle and query if it is a tty.
      */
-    @ikvm.lang.Property(get = "get_handle")
-    private long handle;
-    
     @ikvm.lang.Property(get = "get_fd")
     private int fd;
+
+    @ikvm.lang.Property(get = "get_handle")
+    private long handle;
+
+    private Closeable parent;
+    private List<Closeable> otherParents;
+    private boolean closed;
     
     @cli.System.Security.SecurityCriticalAttribute.Annotation
     private long get_handle()
@@ -118,20 +123,11 @@ public final class FileDescriptor {
     @DllImportAttribute.Annotation("kernel32")
     private static native cli.System.IntPtr GetStdHandle(int nStdHandle);
 
-    /**
-     * A use counter for tracking the FIS/FOS/RAF instances that
-     * use this FileDescriptor. The FIS/FOS.finalize() will not release
-     * the FileDescriptor if it is still under use by any stream.
-     */
-    private AtomicInteger useCount;
-
-
     /**
      * Constructs an (invalid) FileDescriptor
      * object.
      */
     public /**/ FileDescriptor() {
-        useCount = new AtomicInteger();
     }
 
     /**
@@ -271,14 +267,68 @@ public final class FileDescriptor {
         }
     }
 
-    // package private methods used by FIS, FOS and RAF.
+    /*
+     * Package private methods to track referents.
+     * If multiple streams point to the same FileDescriptor, we cycle
+     * through the list of all referents and call close()
+     */
 
-    int incrementAndGetUseCount() {
-        return useCount.incrementAndGet();
+    /**
+     * Attach a Closeable to this FD for tracking.
+     * parent reference is added to otherParents when
+     * needed to make closeAll simpler.
+     */
+    synchronized void attach(Closeable c) {
+        if (parent == null) {
+            // first caller gets to do this
+            parent = c;
+        } else if (otherParents == null) {
+            otherParents = new ArrayList<>();
+            otherParents.add(parent);
+            otherParents.add(c);
+        } else {
+            otherParents.add(c);
+        }
     }
 
-    int decrementAndGetUseCount() {
-        return useCount.decrementAndGet();
+    /**
+     * Cycle through all Closeables sharing this FD and call
+     * close() on each one.
+     *
+     * The caller closeable gets to call close0().
+     */
+    @SuppressWarnings("try")
+    synchronized void closeAll(Closeable releaser) throws IOException {
+        if (!closed) {
+            closed = true;
+            IOException ioe = null;
+            try (Closeable c = releaser) {
+                if (otherParents != null) {
+                    for (Closeable referent : otherParents) {
+                        try {
+                            referent.close();
+                        } catch(IOException x) {
+                            if (ioe == null) {
+                                ioe = x;
+                            } else {
+                                ioe.addSuppressed(x);
+                            }
+                        }
+                    }
+                }
+            } catch(IOException ex) {
+                /*
+                 * If releaser close() throws IOException
+                 * add other exceptions as suppressed.
+                 */
+                if (ioe != null)
+                    ex.addSuppressed(ioe);
+                ioe = ex;
+            } finally {
+                if (ioe != null)
+                    throw ioe;
+            }
+        }
     }
 
     void openReadOnly(String name) throws FileNotFoundException
@@ -325,7 +375,8 @@ public final class FileDescriptor {
         catch (cli.System.UnauthorizedAccessException x3)
         {
             // this is caused by "name" being a directory instead of a file
-            throw new FileNotFoundException(x3.getMessage());
+            // or by name being a read-only file
+            throw new FileNotFoundException(name + " (Access is denied)");
         }
         catch (cli.System.ArgumentException x4)
         {
