@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2007, 2011, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 2007, 2013, Oracle and/or its affiliates. All rights reserved.
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
  * This code is free software; you can redistribute it and/or modify it
@@ -38,7 +38,6 @@ import sun.net.ResourceManager;
  * during socket creation.
  *
  * @author Chris Hegarty
- * @author Jeroen Frijters
  */
 
 class TwoStacksPlainDatagramSocketImpl extends AbstractPlainDatagramSocketImpl
@@ -63,9 +62,30 @@ class TwoStacksPlainDatagramSocketImpl extends AbstractPlainDatagramSocketImpl
      */
     cli.System.Net.Sockets.Socket lastfd=null;
 
+    // true if this socket is exclusively bound
+    private final boolean exclusiveBind;
+
+    /*
+     * Set to true if SO_REUSEADDR is set after the socket is bound to
+     * indicate SO_REUSEADDR is being emulated
+     */
+    private boolean reuseAddressEmulated;
+
+    // emulates SO_REUSEADDR when exclusiveBind is true and socket is bound
+    private boolean isReuseAddress;
+
+    TwoStacksPlainDatagramSocketImpl(boolean exclBind) {
+        exclusiveBind = exclBind;
+    }
+
     protected synchronized void create() throws SocketException {
         fd1 = new FileDescriptor();
-        super.create();
+        try {
+            super.create();
+        } catch (SocketException e) {
+            fd1 = null;
+            throw e;
+        }
     }
 
     protected synchronized void bind(int lport, InetAddress laddr)
@@ -76,6 +96,14 @@ class TwoStacksPlainDatagramSocketImpl extends AbstractPlainDatagramSocketImpl
         }
     }
 
+    @Override
+    protected synchronized void bind0(int lport, InetAddress laddr)
+        throws SocketException
+    {
+        bind0(lport, laddr, exclusiveBind);
+
+    }
+
     protected synchronized void receive(DatagramPacket p)
         throws IOException {
         try {
@@ -91,12 +119,29 @@ class TwoStacksPlainDatagramSocketImpl extends AbstractPlainDatagramSocketImpl
         }
 
         if (optID == SO_BINDADDR) {
-            if (fd != null && fd1 != null) {
+            if ((fd != null && fd1 != null) && !connected) {
                 return anyLocalBoundAddr;
             }
-            return socketGetOption(optID);
-        } else
+            int family = connectedAddress == null ? -1 : connectedAddress.holder().getFamily();
+            return socketLocalAddress(family);
+        } else if (optID == SO_REUSEADDR && reuseAddressEmulated) {
+            return isReuseAddress;
+        } else {
             return super.getOption(optID);
+        }
+    }
+
+    protected void socketSetOption(int opt, Object val)
+        throws SocketException
+    {
+        if (opt == SO_REUSEADDR && exclusiveBind && localPort != 0)  {
+            // socket already bound, emulate
+            reuseAddressEmulated = true;
+            isReuseAddress = (Boolean)val;
+        } else {
+            socketNativeSetOption(opt, val);
+        }
+
     }
 
     protected boolean isClosed() {
@@ -114,9 +159,10 @@ class TwoStacksPlainDatagramSocketImpl extends AbstractPlainDatagramSocketImpl
 
     /* Native methods */
 
-    protected synchronized void bind0(int lport, InetAddress laddr) throws SocketException {
+    protected synchronized void bind0(int lport, InetAddress laddr,
+                                             boolean exclBind) throws SocketException {
         ikvm.internal.JNI.JNIEnv env = new ikvm.internal.JNI.JNIEnv();
-        TwoStacksPlainDatagramSocketImpl_c.bind0(env, this, lport, laddr);
+        TwoStacksPlainDatagramSocketImpl_c.bind0(env, this, lport, laddr, exclBind);
         env.ThrowPendingException();
     }
 
@@ -194,9 +240,9 @@ class TwoStacksPlainDatagramSocketImpl extends AbstractPlainDatagramSocketImpl
         TwoStacksPlainDatagramSocketImpl_c.datagramSocketClose(this);
     }
 
-    protected void socketSetOption(int opt, Object val) throws SocketException {
+    protected void socketNativeSetOption(int opt, Object val) throws SocketException {
         ikvm.internal.JNI.JNIEnv env = new ikvm.internal.JNI.JNIEnv();
-        TwoStacksPlainDatagramSocketImpl_c.socketSetOption(env, this, opt, val);
+        TwoStacksPlainDatagramSocketImpl_c.socketNativeSetOption(env, this, opt, val);
         env.ThrowPendingException();
     }
 
@@ -208,7 +254,7 @@ class TwoStacksPlainDatagramSocketImpl extends AbstractPlainDatagramSocketImpl
     }
 
     protected void connect0(InetAddress address, int port) throws SocketException {
-        if (runningOnMono) {
+        if (ikvm.internal.Util.MONO) {
             // MONOBUG Mono doesn't allow Socket.Connect(IPAddress.Any, 0) to disconnect a datagram socket,
             // so we throw a SocketException, this will cause DatagramSocket to emulate connectedness
             throw new SocketException("connected datagram sockets not supported on Mono");
@@ -218,9 +264,14 @@ class TwoStacksPlainDatagramSocketImpl extends AbstractPlainDatagramSocketImpl
         env.ThrowPendingException();
     }
 
+    protected Object socketLocalAddress(int family) throws SocketException {
+        ikvm.internal.JNI.JNIEnv env = new ikvm.internal.JNI.JNIEnv();
+        Object ret = TwoStacksPlainDatagramSocketImpl_c.socketLocalAddress(env, this, family);
+        env.ThrowPendingException();
+        return ret;
+    }
+
     protected void disconnect0(int family) {
         TwoStacksPlainDatagramSocketImpl_c.disconnect0(this, family);
     }
-
-    private static final boolean runningOnMono = cli.System.Type.GetType("Mono.Runtime") != null;
 }
