@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997, 2007, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 1997, 2013, Oracle and/or its affiliates. All rights reserved.
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
  * This code is free software; you can redistribute it and/or modify it
@@ -25,6 +25,7 @@
 
 package java.net;
 
+import java.io.FileDescriptor;
 import cli.System.Net.IPAddress;
 import cli.System.Net.IPEndPoint;
 import static ikvm.internal.JNI.*;
@@ -270,6 +271,17 @@ final class net_util_md
             optval = tos;
         }
 
+        if (optname == SO_REUSEADDR) {
+            /*
+             * Do not set SO_REUSEADDE if SO_EXCLUSIVEADDUSE is already set
+             */
+            int[] parg = new int[1];
+            rv = NET_GetSockOpt(s, SOL_SOCKET, SO_EXCLUSIVEADDRUSE, parg);
+            if (rv == 0 && parg[0] == 1) {
+                return rv;
+            }
+        }
+
         rv = setsockopt(s, level, optname, optval);
 
         if (rv == SOCKET_ERROR) {
@@ -306,6 +318,11 @@ final class net_util_md
     {
         int rv;
 
+        if (level == IPPROTO_IPV6 && optname == IPV6_TCLASS) {
+            ((int[])optval)[0] = 0;
+            return 0;
+        }
+
         rv = getsockopt(s, level, optname, optval);
 
 
@@ -328,6 +345,18 @@ final class net_util_md
         return rv;
     }
 
+    /*
+     * Sets SO_ECLUSIVEADDRUSE if SO_REUSEADDR is not already set.
+     */
+    static void setExclusiveBind(cli.System.Net.Sockets.Socket fd) {
+        int[] parg = new int[1];
+        int rv = 0;
+        rv = NET_GetSockOpt(fd, SOL_SOCKET, SO_REUSEADDR, parg);
+        if (rv == 0 && parg[0] == 0) {
+            rv = NET_SetSockOpt(fd, SOL_SOCKET, SO_EXCLUSIVEADDRUSE, 1);
+        }
+    }
+
     /*
      * Wrapper for bind winsock call - transparent converts an
      * error related to binding to a port that has exclusive access
@@ -336,7 +365,8 @@ final class net_util_md
      */
     static int NET_Bind(cli.System.Net.Sockets.Socket s, SOCKETADDRESS him)
     {
-        int rv = bind(s, him);
+        int rv;
+        rv = bind(s, him);
 
         if (rv == SOCKET_ERROR) {
             /*
@@ -351,6 +381,17 @@ final class net_util_md
         return rv;
     }
 
+    /*
+     * Wrapper for NET_Bind call. Sets SO_EXCLUSIVEADDRUSE
+     * if required, and then calls NET_BIND
+     */
+    static int NET_WinBind(cli.System.Net.Sockets.Socket s, SOCKETADDRESS him, boolean exclBind)
+    {
+        if (exclBind == JNI_TRUE)
+            setExclusiveBind(s);
+        return NET_Bind(s, him);
+    }
+
     static int NET_SocketClose(cli.System.Net.Sockets.Socket fd) {
         linger l = new linger();
         int ret;
@@ -469,7 +510,7 @@ final class net_util_md
         return SOCKET_ERROR;
     }
 
-    static int NET_BindV6(ipv6bind b) {
+    static int NET_BindV6(ipv6bind b, boolean exclBind) {
         cli.System.Net.Sockets.Socket fd = null;
         cli.System.Net.Sockets.Socket ofd = null;
         int rv;
@@ -486,7 +527,7 @@ final class net_util_md
         if (family == AF_INET && (b.addr.him4.sin_addr.s_addr != INADDR_ANY)) {
             /* bind to v4 only */
             int ret;
-            ret = NET_Bind (b.ipv4_fd, b.addr);
+            ret = NET_WinBind (b.ipv4_fd, b.addr, exclBind);
             if (ret == SOCKET_ERROR) {
                 return CLOSE_SOCKETS_AND_RETURN(fd, ofd, close_fd, close_ofd, b);
             }
@@ -497,7 +538,7 @@ final class net_util_md
         if (family == AF_INET6 && (!IN6ADDR_ISANY(b.addr))) {
             /* bind to v6 only */
             int ret;
-            ret = NET_Bind (b.ipv6_fd, b.addr);
+            ret = NET_WinBind (b.ipv6_fd, b.addr, exclBind);
             if (ret == SOCKET_ERROR) {
                 return CLOSE_SOCKETS_AND_RETURN(fd, ofd, close_fd, close_ofd, b);
             }
@@ -522,7 +563,7 @@ final class net_util_md
             oaddr.set(new IPEndPoint(IPAddress.Any, htons(port)));
         }
 
-        rv = NET_Bind (fd, b.addr);
+        rv = NET_WinBind (fd, b.addr, exclBind);
         if (rv == SOCKET_ERROR) {
             return CLOSE_SOCKETS_AND_RETURN(fd, ofd, close_fd, close_ofd, b);
         }
@@ -567,7 +608,8 @@ final class net_util_md
 
                 /* bind random port on first socket */
                 oaddr.sin_port = 0;
-                rv = NET_Bind (ofd, oaddr);
+                rv = NET_WinBind (ofd, oaddr,
+                                  exclBind);
                 if (rv == SOCKET_ERROR) {
                     return CLOSE_SOCKETS_AND_RETURN(fd, ofd, close_fd, close_ofd, b);
                 }
@@ -582,7 +624,8 @@ final class net_util_md
                 }
                 bound_port = oaddr.sin_port;
                 b.addr.sin_port = bound_port;
-                rv = NET_Bind (fd, b.addr);
+                rv = NET_WinBind (fd, b.addr,
+                                  exclBind);
 
                 if (rv != SOCKET_ERROR) {
                     if (family == AF_INET) {
@@ -607,17 +650,17 @@ final class net_util_md
      * structure for an IPv4 InetAddress.
     */
     static int NET_InetAddressToSockaddr(JNIEnv env, InetAddress iaObj, int port, SOCKETADDRESS him, boolean v4MappedAddress) {
-        if (iaObj.family == InetAddress.IPv4) {
-            him.set(new IPEndPoint(new IPAddress(htonl(iaObj.address) & 0xFFFFFFFFL), port));
+        if (iaObj.holder().family == InetAddress.IPv4) {
+            him.set(new IPEndPoint(new IPAddress(htonl(iaObj.holder().address) & 0xFFFFFFFFL), port));
             return 0;
         } else {
             Inet6Address v6addr = (Inet6Address)iaObj;
             int scope = v6addr.getScopeId();
             if (scope == 0) {
-                him.set(new IPEndPoint(new IPAddress(v6addr.ipaddress), port));
+                him.set(new IPEndPoint(new IPAddress(v6addr.getAddress()), port));
                 return 0;
             } else {
-                him.set(new IPEndPoint(new IPAddress(v6addr.ipaddress, scope & 0xFFFFFFFFL), port));
+                him.set(new IPEndPoint(new IPAddress(v6addr.getAddress(), scope & 0xFFFFFFFFL), port));
                 return 0;
             }
         }
@@ -687,7 +730,7 @@ final class net_util_md
     }
 
     static boolean NET_SockaddrEqualsInetAddress(SOCKETADDRESS him, InetAddress iaObj) {
-        int family = iaObj.family == InetAddress.IPv4 ? AF_INET : AF_INET6;
+        int family = iaObj.holder().family == InetAddress.IPv4 ? AF_INET : AF_INET6;
 
         if (him.sa_family == AF_INET6) {
             byte[] caddrNew = him.him6.sin6_addr;
@@ -698,7 +741,7 @@ final class net_util_md
                         return false;
                     }
                     addrNew = NET_IPv4MappedToIPv4(caddrNew);
-                    addrCur = iaObj.address;
+                    addrCur = iaObj.holder().address;
                     if (addrNew == addrCur) {
                         return true;
                     } else {
@@ -712,7 +755,7 @@ final class net_util_md
                         return false;
                     }
                     scope = ((Inet6Address)iaObj).getScopeId();
-                    caddrCur = ((Inet6Address)iaObj).ipaddress;
+                    caddrCur = ((Inet6Address)iaObj).getAddress();
                     if (NET_IsEqual(caddrNew, caddrCur) && cmpScopeID(scope, him)) {
                         return true;
                     } else {
@@ -725,7 +768,7 @@ final class net_util_md
                 return false;
             }
             addrNew = ntohl(him.him4.sin_addr.s_addr);
-            addrCur = iaObj.address;
+            addrCur = iaObj.holder().address;
             if (addrNew == addrCur) {
                 return true;
             } else {
@@ -734,6 +777,10 @@ final class net_util_md
         }
     }
 
+    static InetAddress NET_SockaddrToInetAddress(JNIEnv env, SOCKETADDRESS him, int[] port) {
+        return NET_SockaddrToInetAddress(him, port);
+    }
+
     static InetAddress NET_SockaddrToInetAddress(SOCKETADDRESS him, int[] port) {
         InetAddress iaObj;
         if (him.sa_family == AF_INET6) {
@@ -741,7 +788,8 @@ final class net_util_md
             if (NET_IsIPv4Mapped(caddr)) {
                 iaObj = new Inet4Address(null, NET_IPv4MappedToIPv4(caddr));
             } else {
-                iaObj = new Inet6Address(null, caddr, getScopeID(him));
+                int scope = getScopeID(him);
+                iaObj = new Inet6Address(null, caddr, scope > 0 ? scope : -1);
             }
             port[0] = ntohs(him.him6.sin_port);
         } else {
@@ -821,4 +869,21 @@ final class net_util_md
     static void Sleep(int ms) {
         cli.System.Threading.Thread.Sleep(ms);
     }
+
+    static cli.System.Net.Sockets.Socket NET_Socket (int domain, int type, int protocol) {
+        cli.System.Net.Sockets.Socket sock;
+        sock = socket (domain, type, protocol);
+        if (sock != INVALID_SOCKET) {
+            //SetHandleInformation((HANDLE)(uintptr_t)sock, HANDLE_FLAG_INHERIT, FALSE);
+        }
+        return sock;
+    }
+
+    static int getInetAddress_addr(JNIEnv env, InetAddress iaObj) {
+        return iaObj.holder().address;
+    }
+
+    static int getInetAddress_family(JNIEnv env, InetAddress iaObj) {
+        return iaObj.holder().family;
+    }
 }
