@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2006-2012 Jeroen Frijters
+  Copyright (C) 2006-2014 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -24,6 +24,7 @@
 
 package sun.misc;
 
+import cli.System.Buffer;
 import cli.System.IntPtr;
 import cli.System.Runtime.InteropServices.Marshal;
 import cli.System.Security.Permissions.SecurityAction;
@@ -31,7 +32,6 @@ import cli.System.Security.Permissions.SecurityPermissionAttribute;
 import ikvm.lang.Internal;
 import java.lang.reflect.Field;
 import java.lang.reflect.Modifier;
-import java.lang.reflect.ReflectHelper;
 import java.security.ProtectionDomain;
 import java.util.ArrayList;
 
@@ -46,20 +46,23 @@ public final class Unsafe
 
     private Unsafe() { }
 
-    @ikvm.internal.HasCallerID
+    @sun.reflect.CallerSensitive
     public static Unsafe getUnsafe()
     {
-        if(ikvm.internal.CallerID.getCallerID().getCallerClassLoader() != null)
+        if(!VM.isSystemDomainLoader(ikvm.internal.CallerID.getCallerID().getCallerClassLoader()))
         {
             throw new SecurityException("Unsafe");
         }
         return theUnsafe;
     }
 
+    private static native Field createFieldAndMakeAccessible(Class c, String field);
+    private static native Field copyFieldAndMakeAccessible(Field field);
+
     // this is the intrinsified version of objectFieldOffset(XXX.class.getDeclaredField("xxx"))
     public long objectFieldOffset(Class c, String field)
     {
-        return fieldOffset(ReflectHelper.createFieldAndMakeAccessible(c, field));
+        return allocateUnsafeFieldId(createFieldAndMakeAccessible(c, field));
     }
 
     // NOTE we have a really lame (and slow) implementation!
@@ -69,13 +72,27 @@ public final class Unsafe
         {
             throw new IllegalArgumentException();
         }
-        return fieldOffset(field);
+        return allocateUnsafeFieldId(field);
+    }
+    
+    public long staticFieldOffset(Field field)
+    {
+        if(!Modifier.isStatic(field.getModifiers()))
+        {
+            throw new IllegalArgumentException();
+        }
+        return allocateUnsafeFieldId(field);
     }
 
     @Deprecated
     public int fieldOffset(Field original)
     {
-        Field copy = ReflectHelper.copyFieldAndMakeAccessible(original);
+        return allocateUnsafeFieldId(original);
+    }
+    
+    static int allocateUnsafeFieldId(Field original)
+    {
+        Field copy = copyFieldAndMakeAccessible(original);
         synchronized(fields)
         {
             int id = fields.size();
@@ -92,11 +109,27 @@ public final class Unsafe
 
     public int arrayIndexScale(Class c)
     {
+        if (c == byte[].class || c == boolean[].class)
+        {
+            return 1;
+        }
+        if (c == char[].class || c == short[].class)
+        {
+            return 2;
+        }
+        if (c == int[].class || c == float[].class)
+        {
+            return 4;
+        }
+        if (c == long[].class || c == double[].class)
+        {
+            return 8;
+        }
         // don't change this, the Unsafe intrinsics depend on this value
         return 1;
     }
 
-    private static Field getField(long offset)
+    static Field getField(long offset)
     {
         synchronized(fields)
         {
@@ -104,43 +137,7 @@ public final class Unsafe
         }
     }
 
-    public boolean compareAndSwapObject(Object obj, long offset, Object expect, Object update)
-    {
-        if(obj instanceof Object[])
-        {
-            Object[] arr = (Object[])obj;
-            int index = (int)offset;
-            synchronized(this)
-            {
-                if(arr[index] == expect)
-                {
-                    arr[index] = update;
-                    return true;
-                }
-                return false;
-            }
-        }
-        else
-        {
-            Field field = getField(offset);
-            synchronized(field)
-            {
-                try
-                {
-                    if(field.get(obj) == expect)
-                    {
-                        field.set(obj, update);
-                        return true;
-                    }
-                    return false;
-                }
-                catch(IllegalAccessException x)
-                {
-                    throw (InternalError)new InternalError().initCause(x);
-                }
-            }
-        }
-    }
+    public final native boolean compareAndSwapObject(Object obj, long offset, Object expect, Object update);
 
     public void putObjectVolatile(Object obj, long offset, Object newValue)
     {
@@ -199,51 +196,22 @@ public final class Unsafe
         }
     }
 
-    public boolean compareAndSwapInt(Object obj, long offset, int expect, int update)
-    {
-        if(obj instanceof int[])
-        {
-            int[] arr = (int[])obj;
-            int index = (int)offset;
-            synchronized(this)
-            {
-                if(arr[index] == expect)
-                {
-                    arr[index] = update;
-                    return true;
-                }
-                return false;
-            }
-        }
-        else
-        {
-            Field field = getField(offset);
-            synchronized(field)
-            {
-                try
-                {
-                    if(field.getInt(obj) == expect)
-                    {
-                        field.setInt(obj, update);
-                        return true;
-                    }
-                    return false;
-                }
-                catch(IllegalAccessException x)
-                {
-                    throw (InternalError)new InternalError().initCause(x);
-                }
-            }
-        }
-    }
+    private static native short ReadInt16(Object obj, long offset);
+    private static native int ReadInt32(Object obj, long offset);
+    private static native long ReadInt64(Object obj, long offset);
+    private static native void WriteInt16(Object obj, long offset, short value);
+    private static native void WriteInt32(Object obj, long offset, int value);
+    private static native void WriteInt64(Object obj, long offset, long value);
+
+    public final native boolean compareAndSwapInt(Object obj, long offset, int expect, int update);
 
     public void putIntVolatile(Object obj, long offset, int newValue)
     {
-        if(obj instanceof int[])
+        if (obj instanceof cli.System.Array)
         {
             synchronized(this)
             {
-                ((int[])obj)[(int)offset] = newValue;
+                WriteInt32(obj, offset, newValue);
             }
         }
         else
@@ -270,11 +238,11 @@ public final class Unsafe
 
     public int getIntVolatile(Object obj, long offset)
     {
-        if(obj instanceof int[])
+        if (obj instanceof cli.System.Array)
         {
             synchronized(this)
             {
-                return ((int[])obj)[(int)offset];
+                return ReadInt32(obj, offset);
             }
         }
         else
@@ -294,51 +262,15 @@ public final class Unsafe
         }
     }
 
-    public boolean compareAndSwapLong(Object obj, long offset, long expect, long update)
-    {
-        if(obj instanceof long[])
-        {
-            long[] arr = (long[])obj;
-            int index = (int)offset;
-            synchronized(this)
-            {
-                if(arr[index] == expect)
-                {
-                    arr[index] = update;
-                    return true;
-                }
-                return false;
-            }
-        }
-        else
-        {
-            Field field = getField(offset);
-            synchronized(field)
-            {
-                try
-                {
-                    if(field.getLong(obj) == expect)
-                    {
-                        field.setLong(obj, update);
-                        return true;
-                    }
-                    return false;
-                }
-                catch(IllegalAccessException x)
-                {
-                    throw (InternalError)new InternalError().initCause(x);
-                }
-            }
-        }
-    }
+    public final native boolean compareAndSwapLong(Object obj, long offset, long expect, long update);
 
     public void putLongVolatile(Object obj, long offset, long newValue)
     {
-        if(obj instanceof long[])
+        if (obj instanceof cli.System.Array)
         {
             synchronized(this)
             {
-                ((long[])obj)[(int)offset] = newValue;
+                WriteInt64(obj, offset, newValue);
             }
         }
         else
@@ -365,11 +297,11 @@ public final class Unsafe
 
     public long getLongVolatile(Object obj, long offset)
     {
-        if(obj instanceof long[])
+        if (obj instanceof cli.System.Array)
         {
             synchronized(this)
             {
-                return ((long[])obj)[(int)offset];
+                return ReadInt64(obj, offset);
             }
         }
         else
@@ -391,9 +323,9 @@ public final class Unsafe
 
     public void putBoolean(Object obj, long offset, boolean newValue)
     {
-        if (obj instanceof boolean[])
+        if (obj instanceof cli.System.Array)
         {
-            ((boolean[])obj)[(int)offset] = newValue;
+            Buffer.SetByte((cli.System.Array)obj, (int)offset, newValue ? (byte)1 : (byte)0);
         }
         else
         {
@@ -408,11 +340,16 @@ public final class Unsafe
         }
     }
 
+    public void putBooleanVolatile(Object obj, long offset, boolean newValue)
+    {
+        putBoolean(obj, offset, newValue);
+    }
+
     public boolean getBoolean(Object obj, long offset)
     {
-        if (obj instanceof boolean[])
+        if (obj instanceof cli.System.Array)
         {
-            return ((boolean[])obj)[(int)offset];
+            return Buffer.GetByte((cli.System.Array)obj, (int)offset) != 0;
         }
         else
         {
@@ -427,11 +364,16 @@ public final class Unsafe
         }
     }
 
+    public boolean getBooleanVolatile(Object obj, long offset)
+    {
+        return getBoolean(obj, offset);
+    }
+
     public void putByte(Object obj, long offset, byte newValue)
     {
-        if (obj instanceof byte[])
+        if (obj instanceof cli.System.Array)
         {
-            ((byte[])obj)[(int)offset] = newValue;
+            Buffer.SetByte((cli.System.Array)obj, (int)offset, newValue);
         }
         else
         {
@@ -446,11 +388,16 @@ public final class Unsafe
         }
     }
 
+    public void putByteVolatile(Object obj, long offset, byte newValue)
+    {
+        putByte(obj, offset, newValue);
+    }
+
     public byte getByte(Object obj, long offset)
     {
-        if (obj instanceof byte[])
+        if (obj instanceof cli.System.Array)
         {
-            return ((byte[])obj)[(int)offset];
+            return Buffer.GetByte((cli.System.Array)obj, (int)offset);
         }
         else
         {
@@ -465,11 +412,16 @@ public final class Unsafe
         }
     }
 
+    public byte getByteVolatile(Object obj, long offset)
+    {
+        return getByte(obj, offset);
+    }
+
     public void putChar(Object obj, long offset, char newValue)
     {
-        if (obj instanceof char[])
+        if (obj instanceof cli.System.Array)
         {
-            ((char[])obj)[(int)offset] = newValue;
+            WriteInt16(obj, offset, (short)newValue);
         }
         else
         {
@@ -484,11 +436,16 @@ public final class Unsafe
         }
     }
 
+    public void putCharVolatile(Object obj, long offset, char newValue)
+    {
+        putChar(obj, offset, newValue);
+    }
+
     public char getChar(Object obj, long offset)
     {
-        if (obj instanceof char[])
+        if (obj instanceof cli.System.Array)
         {
-            return ((char[])obj)[(int)offset];
+            return (char)ReadInt16(obj, offset);
         }
         else
         {
@@ -503,11 +460,16 @@ public final class Unsafe
         }
     }
 
+    public char getCharVolatile(Object obj, long offset)
+    {
+        return getChar(obj, offset);
+    }
+
     public void putShort(Object obj, long offset, short newValue)
     {
-        if (obj instanceof short[])
+        if (obj instanceof cli.System.Array)
         {
-            ((short[])obj)[(int)offset] = newValue;
+            WriteInt16(obj, offset, newValue);
         }
         else
         {
@@ -522,11 +484,16 @@ public final class Unsafe
         }
     }
 
+    public void putShortVolatile(Object obj, long offset, short newValue)
+    {
+        putShort(obj, offset, newValue);
+    }
+
     public short getShort(Object obj, long offset)
     {
-        if (obj instanceof short[])
+        if (obj instanceof cli.System.Array)
         {
-            return ((short[])obj)[(int)offset];
+            return ReadInt16(obj, offset);
         }
         else
         {
@@ -541,11 +508,16 @@ public final class Unsafe
         }
     }
 
+    public short getShortVolatile(Object obj, long offset)
+    {
+        return getShort(obj, offset);
+    }
+
     public void putInt(Object obj, long offset, int newValue)
     {
-        if (obj instanceof int[])
+        if (obj instanceof cli.System.Array)
         {
-            ((int[])obj)[(int)offset] = newValue;
+            WriteInt32(obj, offset, newValue);
         }
         else
         {
@@ -562,9 +534,9 @@ public final class Unsafe
 
     public int getInt(Object obj, long offset)
     {
-        if (obj instanceof int[])
+        if (obj instanceof cli.System.Array)
         {
-            return ((int[])obj)[(int)offset];
+            return ReadInt32(obj, offset);
         }
         else
         {
@@ -581,9 +553,9 @@ public final class Unsafe
 
     public void putFloat(Object obj, long offset, float newValue)
     {
-        if (obj instanceof float[])
+        if (obj instanceof cli.System.Array)
         {
-            ((float[])obj)[(int)offset] = newValue;
+            WriteInt32(obj, offset, Float.floatToRawIntBits(newValue));
         }
         else
         {
@@ -598,11 +570,16 @@ public final class Unsafe
         }
     }
 
+    public void putFloatVolatile(Object obj, long offset, float newValue)
+    {
+        putFloat(obj, offset, newValue);
+    }
+
     public float getFloat(Object obj, long offset)
     {
-        if (obj instanceof float[])
+        if (obj instanceof cli.System.Array)
         {
-            return ((float[])obj)[(int)offset];
+            return Float.intBitsToFloat(ReadInt32(obj, offset));
         }
         else
         {
@@ -617,11 +594,16 @@ public final class Unsafe
         }
     }
 
+    public float getFloatVolatile(Object obj, long offset)
+    {
+        return getFloat(obj, offset);
+    }
+
     public void putLong(Object obj, long offset, long newValue)
     {
-        if (obj instanceof long[])
+        if (obj instanceof cli.System.Array)
         {
-            ((long[])obj)[(int)offset] = newValue;
+            WriteInt64(obj, offset, newValue);
         }
         else
         {
@@ -638,9 +620,9 @@ public final class Unsafe
 
     public long getLong(Object obj, long offset)
     {
-        if (obj instanceof long[])
+        if (obj instanceof cli.System.Array)
         {
-            return ((long[])obj)[(int)offset];
+            return ReadInt64(obj, offset);
         }
         else
         {
@@ -657,9 +639,9 @@ public final class Unsafe
 
     public void putDouble(Object obj, long offset, double newValue)
     {
-        if (obj instanceof double[])
+        if (obj instanceof cli.System.Array)
         {
-            ((double[])obj)[(int)offset] = newValue;
+            WriteInt64(obj, offset, Double.doubleToRawLongBits(newValue));
         }
         else
         {
@@ -674,11 +656,19 @@ public final class Unsafe
         }
     }
 
+    public void putDoubleVolatile(Object obj, long offset, double newValue)
+    {
+        synchronized (this)
+        {
+            putDouble(obj, offset, newValue);
+        }
+    }
+
     public double getDouble(Object obj, long offset)
     {
-        if (obj instanceof double[])
+        if (obj instanceof cli.System.Array)
         {
-            return ((double[])obj)[(int)offset];
+            return Double.longBitsToDouble(ReadInt64(obj, offset));
         }
         else
         {
@@ -693,6 +683,14 @@ public final class Unsafe
         }
     }
 
+    public double getDoubleVolatile(Object obj, long offset)
+    {
+        synchronized (this)
+        {
+            return getDouble(obj, offset);
+        }
+    }
+
     public void putObject(Object obj, long offset, Object newValue)
     {
         if (obj instanceof Object[])
@@ -863,6 +861,10 @@ public final class Unsafe
     @cli.System.Security.SecurityCriticalAttribute.Annotation
     public long allocateMemory(long bytes)
     {
+        if (bytes == 0)
+        {
+            return 0;
+        }
         try
         {
             if (false) throw new cli.System.OutOfMemoryException();
@@ -874,6 +876,26 @@ public final class Unsafe
         }
     }
 
+    @SecurityPermissionAttribute.Annotation(value = SecurityAction.__Enum.LinkDemand, UnmanagedCode = true)
+    @cli.System.Security.SecurityCriticalAttribute.Annotation
+    public long reallocateMemory(long address, long bytes)
+    {
+        if (bytes == 0)
+        {
+            freeMemory(address);
+            return 0;
+        }
+        try
+        {
+            if (false) throw new cli.System.OutOfMemoryException();
+            return Marshal.ReAllocHGlobal(IntPtr.op_Explicit(address), IntPtr.op_Explicit(bytes)).ToInt64();
+        }
+        catch (cli.System.OutOfMemoryException x)
+        {
+            throw new OutOfMemoryError(x.get_Message());
+        }
+    }
+
     @SecurityPermissionAttribute.Annotation(value = SecurityAction.__Enum.LinkDemand, UnmanagedCode = true)
     @cli.System.Security.SecurityCriticalAttribute.Annotation
     public void freeMemory(long address)
@@ -1122,6 +1144,10 @@ public final class Unsafe
         }
         else
         {
+            if (time == 0)
+            {
+                time = Long.MAX_VALUE;
+            }
             java.util.concurrent.locks.LockSupport.parkNanos(time);
         }
     }
@@ -1135,9 +1161,23 @@ public final class Unsafe
     {
         return null;
     }
+    
+    @Deprecated
+    public Object staticFieldBase(Class<?> c)
+    {
+        return null;
+    }
+
+    public native boolean shouldBeInitialized(Class<?> c);
 
     public native Class defineClass(String name, byte[] buf, int offset, int length, ClassLoader cl, ProtectionDomain pd);
 
+    @Deprecated
+    @sun.reflect.CallerSensitive
+    public native Class defineClass(String name, byte[] b, int off, int len);
+
+    public native Class defineAnonymousClass(Class hostClass, byte[] data, Object[] cpPatches);
+
     public void monitorEnter(Object o)
     {
         cli.System.Threading.Monitor.Enter(o);
@@ -1152,4 +1192,79 @@ public final class Unsafe
     {
         return cli.System.Threading.Monitor.TryEnter(o);
     }
+
+    public final int getAndAddInt(Object o, long offset, int delta)
+    {
+        for (;;)
+        {
+            int value = getIntVolatile(o, offset);
+            if (compareAndSwapInt(o, offset, value, value + delta))
+            {
+                return value;
+            }
+        }
+    }
+
+    public final long getAndAddLong(Object o, long offset, long delta)
+    {
+        for (;;)
+        {
+            long value = getLongVolatile(o, offset);
+            if (compareAndSwapLong(o, offset, value, value + delta))
+            {
+                return value;
+            }
+        }
+    }
+
+    public final int getAndSetInt(Object o, long offset, int newValue)
+    {
+        for (;;)
+        {
+            int value = getIntVolatile(o, offset);
+            if (compareAndSwapInt(o, offset, value, newValue))
+            {
+                return value;
+            }
+        }
+    }
+
+    public final long getAndSetLong(Object o, long offset, long newValue)
+    {
+        for (;;)
+        {
+            long value = getLongVolatile(o, offset);
+            if (compareAndSwapLong(o, offset, value, newValue))
+            {
+                return value;
+            }
+        }
+    }
+
+    public final Object getAndSetObject(Object o, long offset, Object newValue)
+    {
+        for (;;)
+        {
+            Object value = getObjectVolatile(o, offset);
+            if (compareAndSwapObject(o, offset, value, newValue))
+            {
+                return value;
+            }
+        }
+    }
+
+    public void loadFence()
+    {
+        cli.System.Threading.Thread.MemoryBarrier();
+    }
+
+    public void storeFence()
+    {
+        cli.System.Threading.Thread.MemoryBarrier();
+    }
+
+    public void fullFence()
+    {
+        cli.System.Threading.Thread.MemoryBarrier();
+    }
 }
