@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1996, 2011, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 1996, 2014, Oracle and/or its affiliates. All rights reserved.
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
  * This code is free software; you can redistribute it and/or modify it
@@ -48,6 +48,7 @@ public class VM {
         return suspended;
     }
 
+    @SuppressWarnings("deprecation")
     public static boolean allowThreadSuspension(ThreadGroup g, boolean b) {
         return g.allowThreadSuspension(b);
     }
@@ -145,10 +146,45 @@ public class VM {
      */
     // public native static void writeJavaProfilerReport();
 
+
+    private static volatile boolean booted = false;
+    private static final Object lock = new Object();
+
+    static {
+        // [IKVM] force System properties initialization ("booting")
+        System.lineSeparator();
+    }
+
+    // Invoked by by System.initializeSystemClass just before returning.
+    // Subsystems that are invoked during initialization can check this
+    // property in order to avoid doing things that should wait until the
+    // application class loader has been set up.
+    //
+    // [IKVM] The above isn't applicable. We only use the booted flag
+    // for the system properties (this is required because the system properties
+    // use java.util.Hashtable (via java.util.Properties) and it relies on
+    // the booted flag to determine whether it is safe to query the system
+    // properties).
+    public static void booted() {
+        synchronized (lock) {
+            booted = true;
+            lock.notifyAll();
+        }
+    }
+
     public static boolean isBooted() {
-        // [IKVM] we support arbitrary order class initialization,
-        //  so we don't distinguish between booted and not booted states
-        return true;
+        return booted;
+    }
+
+    // Waits until VM completes initialization
+    //
+    // This method is invoked by the Finalizer thread
+    public static void awaitBooted() throws InterruptedException {
+        synchronized (lock) {
+            while (!booted) {
+                lock.wait();
+            }
+        }
     }
 
     // Returns the maximum amount of allocatable direct buffer memory.
@@ -197,6 +233,14 @@ public class VM {
         return allowArraySyntax;
     }
 
+    /**
+     * Returns true if the given class loader is in the system domain
+     * in which all permissions are granted.
+     */
+    public static boolean isSystemDomainLoader(ClassLoader loader) {
+        return loader == null;
+    }
+
     /**
      * Returns the system property of the specified key saved at
      * system initialization time.  This method should only be used
@@ -210,27 +254,28 @@ public class VM {
      *
      */
     public static String getSavedProperty(String key) {
-        // [IKVM] force System properties initialization
-        System.lineSeparator();
-
-        if (savedProps.isEmpty())
+        if (Lazy.savedProps.isEmpty())
             throw new IllegalStateException("Should be non-empty if initialized");
 
-        return savedProps.getProperty(key);
+        return Lazy.savedProps.getProperty(key);
     }
 
     // TODO: the Property Management needs to be refactored and
     // the appropriate prop keys need to be accessible to the
     // calling classes to avoid duplication of keys.
-    private static final Properties savedProps = new Properties();
+    static final class Lazy {
+        static final Properties savedProps = new Properties();
+    }
 
     // Save a private copy of the system properties and remove
     // the system properties that are not intended for public access.
     //
     // This method can only be invoked during system initialization.
     public static void saveAndRemoveProperties(Properties props) {
+        if (booted)
+            throw new IllegalStateException("System initialization has completed");
 
-        savedProps.putAll(props);
+        Lazy.savedProps.putAll(props);
 
         // Set the maximum amount of direct memory.  This value is controlled
         // by the vm option -XX:MaxDirectMemorySize=<size>.
@@ -262,6 +307,9 @@ public class VM {
 
         // used by sun.launcher.LauncherHelper
         props.remove("sun.java.launcher.diag");
+
+        // used by sun.misc.URLClassPath
+        props.remove("sun.cds.enableSharedLookupCache");
     }
 
     // Initialize any miscellenous operating system settings that need to be
@@ -341,6 +389,12 @@ public class VM {
     private final static int JVMTI_THREAD_STATE_WAITING_INDEFINITELY = 0x0010;
     private final static int JVMTI_THREAD_STATE_WAITING_WITH_TIMEOUT = 0x0020;
 
+    /*
+     * Returns the first non-null class loader up the execution stack,
+     * or null if only code from the null class loader is on the stack.
+     */
+    public static native ClassLoader latestUserDefinedLoader();
+
     static {
         initialize();
     }
