@@ -36,6 +36,8 @@ import cli.System.IO.FileMode;
 import cli.System.IO.FileShare;
 import cli.System.IO.FileStream;
 import cli.System.IO.FileOptions;
+import cli.System.Runtime.InteropServices.DllImportAttribute;
+import cli.System.Runtime.InteropServices.Marshal;
 import cli.System.Security.AccessControl.FileSystemRights;
 import com.sun.nio.file.ExtendedOpenOption;
 import java.io.FileDescriptor;
@@ -60,7 +62,7 @@ final class NetFileSystemProvider extends AbstractFileSystemProvider
     private final NetFileSystem fs = new NetFileSystem(this);
     private final HashMap<String, FileStore> stores = new HashMap<String, FileStore>();
 
-    final synchronized FileStore getFileStore(DriveInfo drive)
+    final synchronized FileStore getFileStore(DriveInfo drive) throws IOException
     {
         String name = drive.get_Name().toLowerCase();
         FileStore fs = stores.get(name);
@@ -338,7 +340,7 @@ final class NetFileSystemProvider extends AbstractFileSystemProvider
             }
         }
 
-        return FileChannelImpl.open(open(npath.path, mode, rights, share, options), read, write, append, null);
+        return FileChannelImpl.open(open(npath.path, mode, rights, share, options), npath.path, read, write, append, null);
     }
 
     private static FileDescriptor open(String path, int mode, int rights, int share, int options) throws IOException
@@ -628,6 +630,7 @@ final class NetFileSystemProvider extends AbstractFileSystemProvider
         NetPath nsource = NetPath.from(source);
         NetPath ntarget = NetPath.from(target);
         boolean overwrite = false;
+        boolean atomicMove = false;
         for (CopyOption opt : options)
         {
             if (opt == StandardCopyOption.REPLACE_EXISTING)
@@ -636,7 +639,14 @@ final class NetFileSystemProvider extends AbstractFileSystemProvider
             }
             else if (opt == StandardCopyOption.ATOMIC_MOVE)
             {
-                throw new AtomicMoveNotSupportedException(nsource.path, ntarget.path, "Unsupported copy option");
+                if (WINDOWS)
+                {
+                    atomicMove = true;
+                }
+                else
+                {
+                    throw new AtomicMoveNotSupportedException(nsource.path, ntarget.path, "Unsupported copy option");
+                }
             }
             else
             {
@@ -651,6 +661,36 @@ final class NetFileSystemProvider extends AbstractFileSystemProvider
             sm.checkRead(nsource.path);
             sm.checkWrite(ntarget.path);
         }
+        if (atomicMove)
+        {
+            int MOVEFILE_REPLACE_EXISTING = 1;
+            if (MoveFileEx(nsource.path, ntarget.path, MOVEFILE_REPLACE_EXISTING) == 0)
+            {
+                final int ERROR_FILE_NOT_FOUND = 2;
+                final int ERROR_PATH_NOT_FOUND = 3;
+                final int ERROR_ACCESS_DENIED = 5;
+                final int ERROR_NOT_SAME_DEVICE = 17;
+                final int ERROR_FILE_EXISTS = 80;
+                final int ERROR_ALREADY_EXISTS = 183;
+                int lastError = Marshal.GetLastWin32Error();
+                switch (lastError)
+                {
+                    case ERROR_FILE_NOT_FOUND:
+                    case ERROR_PATH_NOT_FOUND:
+                        throw new NoSuchFileException(nsource.path, ntarget.path, null);
+                    case ERROR_ACCESS_DENIED:
+                        throw new AccessDeniedException(nsource.path, ntarget.path, null);
+                    case ERROR_NOT_SAME_DEVICE:
+                        throw new AtomicMoveNotSupportedException(nsource.path, ntarget.path, "Unsupported copy option");
+                    case ERROR_FILE_EXISTS:
+                    case ERROR_ALREADY_EXISTS:
+                        throw new FileAlreadyExistsException(nsource.path, ntarget.path, null);
+                    default:
+                        throw new FileSystemException(nsource.path, ntarget.path, "Error " + lastError);
+                }
+            }
+            return;
+        }
         try
         {
             if (false) throw new cli.System.ArgumentException();
@@ -711,6 +751,9 @@ final class NetFileSystemProvider extends AbstractFileSystemProvider
         }
     }
 
+    @DllImportAttribute.Annotation(value="kernel32", SetLastError=true)
+    private static native int MoveFileEx(String lpExistingFileName, String lpNewFileName, int dwFlags);
+
     public boolean isSameFile(Path path, Path path2) throws IOException
     {
         if (path.equals(path2))
@@ -754,10 +797,22 @@ final class NetFileSystemProvider extends AbstractFileSystemProvider
     private static class NetFileStore extends FileStore
     {
         private final DriveInfo info;
+        private final String name;
+        private final String type;
 
-        NetFileStore(DriveInfo info)
+        NetFileStore(DriveInfo info) throws IOException
         {
             this.info = info;
+            try
+            {
+                if (false) throw new cli.System.IO.IOException();
+                name = info.get_VolumeLabel();
+                type = info.get_DriveFormat();
+            }
+            catch (cli.System.IO.IOException x)
+            {
+                throw new IOException(x.getMessage());
+            }
         }
 
         public Object getAttribute(String attribute) throws IOException
@@ -782,17 +837,41 @@ final class NetFileSystemProvider extends AbstractFileSystemProvider
 
         public long getTotalSpace() throws IOException
         {
-            return info.get_TotalSize();
+            try
+            {
+                if (false) throw new cli.System.IO.IOException();
+                return info.get_TotalSize();
+            }
+            catch (cli.System.IO.IOException x)
+            {
+                throw new IOException(x.getMessage());
+            }
         }
 
         public long getUnallocatedSpace() throws IOException
         {
-            return info.get_TotalFreeSpace();
+            try
+            {
+                if (false) throw new cli.System.IO.IOException();
+                return info.get_TotalFreeSpace();
+            }
+            catch (cli.System.IO.IOException x)
+            {
+                throw new IOException(x.getMessage());
+            }
         }
 
         public long getUsableSpace() throws IOException
         {
-            return info.get_AvailableFreeSpace();
+            try
+            {
+                if (false) throw new cli.System.IO.IOException();
+                return info.get_AvailableFreeSpace();
+            }
+            catch (cli.System.IO.IOException x)
+            {
+                throw new IOException(x.getMessage());
+            }
         }
 
         public boolean isReadOnly()
@@ -802,7 +881,7 @@ final class NetFileSystemProvider extends AbstractFileSystemProvider
 
         public String name()
         {
-            return info.get_VolumeLabel();
+            return name;
         }
 
         public boolean supportsFileAttributeView(Class<? extends FileAttributeView> type)
@@ -819,8 +898,13 @@ final class NetFileSystemProvider extends AbstractFileSystemProvider
 
         public String type()
         {
-            return info.get_DriveFormat();
+            return type;
         }        
+
+        public String toString()
+        {
+            return name + " (" + info.get_Name().charAt(0) + ":)";
+        }
     }
 
     public FileStore getFileStore(Path path) throws IOException
@@ -1063,7 +1147,7 @@ final class NetFileSystemProvider extends AbstractFileSystemProvider
 
             public long size()
             {
-                return info.get_Length();
+                return info.get_Exists() ? info.get_Length() : 0;
             }
 
             public boolean isArchive()
@@ -1099,7 +1183,15 @@ final class NetFileSystemProvider extends AbstractFileSystemProvider
                 if (false) throw new cli.System.ArgumentException();
                 if (false) throw new cli.System.IO.FileNotFoundException();
                 if (false) throw new cli.System.IO.IOException();
-                return new DosFileAttributesImpl(new FileInfo(path));
+                FileInfo info = new FileInfo(path);
+                // We have to rely on the (undocumented) fact that FileInfo.Attributes returns -1
+                // when the path does not exist. We need this to work for both files and directories
+                // and this is the only efficient way to do that.
+                if (info.get_Attributes().Value == -1)
+                {
+                    throw new NoSuchFileException(path);
+                }
+                return new DosFileAttributesImpl(info);
             }
             catch (cli.System.IO.FileNotFoundException _)
             {
@@ -1152,6 +1244,10 @@ final class NetFileSystemProvider extends AbstractFileSystemProvider
                     info.set_Attributes(cli.System.IO.FileAttributes.wrap(info.get_Attributes().Value & ~attr));
                 }
             }
+            catch (cli.System.IO.FileNotFoundException _)
+            {
+                throw new NoSuchFileException(path);
+            }
             catch (cli.System.ArgumentException
                  | cli.System.IO.IOException x)
             {
