@@ -1,27 +1,30 @@
 /*
-  Copyright (C) 2009 Volker Berlin (i-net software)
-  Copyright (C) 2010, 2011 Karsten Heinrich (i-net software)
-
-  This software is provided 'as-is', without any express or implied
-  warranty.  In no event will the authors be held liable for any damages
-  arising from the use of this software.
-
-  Permission is granted to anyone to use this software for any purpose,
-  including commercial applications, and to alter it and redistribute it
-  freely, subject to the following restrictions:
-
-  1. The origin of this software must not be misrepresented; you must not
-     claim that you wrote the original software. If you use this software
-     in a product, an acknowledgment in the product documentation would be
-     appreciated but is not required.
-  2. Altered source versions must be plainly marked as such, and must not be
-     misrepresented as being the original software.
-  3. This notice may not be removed or altered from any source distribution.
-
-  Jeroen Frijters
-  jeroen@frijters.net
-  
+ * Copyright (c) 2000, 2013, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (C) 2009, 2012 Volker Berlin (i-net software)
+ * Copyright (C) 2010, 2011 Karsten Heinrich (i-net software)
+ * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
+ *
+ * This code is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU General Public License version 2 only, as
+ * published by the Free Software Foundation.  Oracle designates this
+ * particular file as subject to the "Classpath" exception as provided
+ * by Oracle in the LICENSE file that accompanied this code.
+ *
+ * This code is distributed in the hope that it will be useful, but WITHOUT
+ * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+ * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
+ * version 2 for more details (a copy is included in the LICENSE file that
+ * accompanied this code).
+ *
+ * You should have received a copy of the GNU General Public License version
+ * 2 along with this work; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
+ *
+ * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
+ * or visit www.oracle.com if you need additional information or have any
+ * questions.
  */
+
 package sun.print;
 
 import java.awt.Color;
@@ -326,6 +329,10 @@ public class Win32PrintJob implements CancelablePrintJob{
             throw new PrintException("can't get print data: " + e.toString());
         }
 
+        if (data == null) {
+            throw new PrintException("Null print data.");
+        }
+
         if(flavor == null || (!service.isDocFlavorSupported(flavor))){
             notifyEvent(PrintJobEvent.JOB_FAILED);
             throw new PrintJobFlavorException("invalid flavor", flavor);
@@ -906,22 +913,21 @@ public class Win32PrintJob implements CancelablePrintJob{
                         return;
                     }
                 }
-                Graphics printGraphics = peer.createGraphics(ev.get_Graphics());
-                if( printGraphics instanceof Graphics2D ){
-                	Graphics2D g2d = ((Graphics2D)printGraphics);
-                	int tX = (int) pageFormat.getWidth(); 
-                	int tY = (int) pageFormat.getHeight();
-                	// apply Java to .NET scaling (1/72 inch to 1/100 inch)
-                	g2d.scale(100d/72d, 100d/72d);
-                	// NOTE on Landscape printing:
-                	// Setting landscape to true on the printer settings
-                	// of a page already rotates the page! The orig. java code rotates the page itself,
-                	// for .NET this is not required.
-                	if( orient == OrientationRequested.REVERSE_LANDSCAPE){
-                		g2d.translate( tX, tY );
-                		g2d.rotate( Math.PI );
-                	}
-                }
+                Graphics2D printGraphics = peer.createGraphics(ev.get_Graphics());
+            	Graphics2D g2d = ((Graphics2D)printGraphics);
+            	int tX = (int) pageFormat.getWidth(); 
+            	int tY = (int) pageFormat.getHeight();
+            	// apply Java to .NET scaling (1/72 inch to 1/100 inch)
+            	g2d.scale(100d/72d, 100d/72d);
+            	// NOTE on Landscape printing:
+            	// Setting landscape to true on the printer settings
+            	// of a page already rotates the page! The orig. java code rotates the page itself,
+            	// for .NET this is not required.
+            	if( orient == OrientationRequested.REVERSE_LANDSCAPE){
+            		g2d.translate( tX, tY );
+            		g2d.rotate( Math.PI );
+            	}
+
                 printable.print(printGraphics, pageFormat, realPage);
                 
                 realPage = pageRanges.getPageForIndex(++pageIndex);
@@ -929,15 +935,7 @@ public class Win32PrintJob implements CancelablePrintJob{
 	                printable = pageable.getPrintable(realPage);
 	                pageFormat = pageable.getPageFormat(realPage);
 	                int pageResult = printable.print(peekGraphics, pageFormat, realPage);
-	                if( pageResult == Printable.PAGE_EXISTS ){
-	                	ev.set_HasMorePages( true );
-	                } else {
-	                	if( pageRanges.checkJobComplete(pageIndex) ){
-	                		ev.set_HasMorePages( false );
-	                	} else {
-	                		ev.set_HasMorePages( true );
-	                	}
-	                }
+                	ev.set_HasMorePages( pageResult == Printable.PAGE_EXISTS );
                 } else {
                 	ev.set_HasMorePages( false );
                 }
