@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2008-2012 Jeroen Frijters
+  Copyright (C) 2008-2015 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -25,7 +25,9 @@ using System;
 using System.Collections.Generic;
 using System.IO;
 using System.Diagnostics;
+#if !NO_SYMBOL_WRITER
 using System.Diagnostics.SymbolStore;
+#endif
 using System.Security.Cryptography;
 using System.Resources;
 using System.Runtime.CompilerServices;
@@ -39,7 +41,8 @@ namespace IKVM.Reflection.Emit
 	public sealed class ModuleBuilder : Module, ITypeOwner
 	{
 		private static readonly bool usePublicKeyAssemblyReference = false;
-		private Guid mvid = Guid.NewGuid();
+		private Guid mvid;
+		private uint timestamp;
 		private long imageBaseAddress = 0x00400000;
 		private long stackReserve = -1;
 		private int fileAlignment = 0x200;
@@ -55,7 +58,6 @@ namespace IKVM.Reflection.Emit
 		internal readonly ByteBuffer methodBodies = new ByteBuffer(128 * 1024);
 		internal readonly List<int> tokenFixupOffsets = new List<int>();
 		internal readonly ByteBuffer initializedData = new ByteBuffer(512);
-		internal readonly ByteBuffer manifestResources = new ByteBuffer(512);
 		internal ResourceSection unmanagedResources;
 		private readonly Dictionary<MemberRefKey, int> importedMemberRefs = new Dictionary<MemberRefKey, int>();
 		private readonly Dictionary<MethodSpecKey, int> importedMethodSpecs = new Dictionary<MethodSpecKey, int>();
@@ -71,30 +73,81 @@ namespace IKVM.Reflection.Emit
 		internal readonly List<VTableFixups> vtablefixups = new List<VTableFixups>();
 		internal readonly List<UnmanagedExport> unmanagedExports = new List<UnmanagedExport>();
 		private List<InterfaceImplCustomAttribute> interfaceImplCustomAttributes;
-		private List<ResourceWriterRecord> resourceWriters;
+		private readonly List<ResourceWriterRecord> resourceWriters = new List<ResourceWriterRecord>();
 		private bool saved;
 
 		private struct ResourceWriterRecord
 		{
 			private readonly string name;
+#if !CORECLR
 			private readonly ResourceWriter rw;
-			private readonly MemoryStream mem;
+#endif
+			private readonly Stream stream;
 			private readonly ResourceAttributes attributes;
 
-			internal ResourceWriterRecord(string name, ResourceWriter rw, MemoryStream mem, ResourceAttributes attributes)
+#if CORECLR
+			internal ResourceWriterRecord(string name, Stream stream, ResourceAttributes attributes)
+			{
+				this.name = name;
+				this.stream = stream;
+				this.attributes = attributes;
+			}
+#else
+			internal ResourceWriterRecord(string name, Stream stream, ResourceAttributes attributes)
+				: this(name, null, stream, attributes)
+			{
+			}
+
+			internal ResourceWriterRecord(string name, ResourceWriter rw, Stream stream, ResourceAttributes attributes)
 			{
 				this.name = name;
 				this.rw = rw;
-				this.mem = mem;
+				this.stream = stream;
 				this.attributes = attributes;
 			}
+#endif
 
-			internal void Emit(ModuleBuilder mb)
+			internal void Emit(ModuleBuilder mb, int offset)
 			{
-				rw.Generate();
-				mem.Position = 0;
-				mb.DefineManifestResource(name, mem, attributes);
-				rw.Close();
+#if !CORECLR
+				if (rw != null)
+				{
+					rw.Generate();
+				}
+#endif
+				ManifestResourceTable.Record rec = new ManifestResourceTable.Record();
+				rec.Offset = offset;
+				rec.Flags = (int)attributes;
+				rec.Name = mb.Strings.Add(name);
+				rec.Implementation = 0;
+				mb.ManifestResource.AddRecord(rec);
+			}
+
+			internal int GetLength()
+			{
+				return 4 + (int)stream.Length;
+			}
+
+			internal void Write(MetadataWriter mw)
+			{
+				mw.Write((int)stream.Length);
+				stream.Position = 0;
+				byte[] buffer = new byte[8192];
+				int length;
+				while ((length = stream.Read(buffer, 0, buffer.Length)) != 0)
+				{
+					mw.Write(buffer, 0, length);
+				}
+			}
+
+			internal void Close()
+			{
+#if !CORECLR
+				if (rw != null)
+				{
+					rw.Close();
+				}
+#endif
 			}
 		}
 
@@ -140,7 +193,7 @@ namespace IKVM.Reflection.Emit
 			public override bool Equals(object obj)
 			{
 				MemberRefKey? other = obj as MemberRefKey?;
-				return other != null && Equals(other);
+				return other != null && Equals(other.Value);
 			}
 
 			public override int GetHashCode()
@@ -180,7 +233,7 @@ namespace IKVM.Reflection.Emit
 			public override bool Equals(object obj)
 			{
 				MethodSpecKey? other = obj as MethodSpecKey?;
-				return other != null && Equals(other);
+				return other != null && Equals(other.Value);
 			}
 
 			public override int GetHashCode()
@@ -198,6 +251,15 @@ namespace IKVM.Reflection.Emit
 			if (emitSymbolInfo)
 			{
 				symbolWriter = SymbolSupport.CreateSymbolWriterFor(this);
+				if (universe.Deterministic && !symbolWriter.IsDeterministic)
+				{
+					throw new NotSupportedException();
+				}
+			}
+			if (!universe.Deterministic)
+			{
+				__PEHeaderTimeDateStamp = DateTime.UtcNow;
+				mvid = Guid.NewGuid();
 			}
 			// <Module> must be the first record in the TypeDef table
 			moduleType = new TypeBuilder(this, null, "<Module>");
@@ -384,16 +446,16 @@ namespace IKVM.Reflection.Emit
 			moduleType.CreateType();
 		}
 
-		internal void AddTypeForwarder(Type type)
+		internal void AddTypeForwarder(Type type, bool includeNested)
 		{
 			ExportType(type);
-			if (!type.__IsMissing)
+			if (includeNested && !type.__IsMissing)
 			{
 				foreach (Type nested in type.GetNestedTypes(BindingFlags.Public | BindingFlags.NonPublic))
 				{
 					// we export all nested types (i.e. even the private ones)
 					// (this behavior is the same as the C# compiler)
-					AddTypeForwarder(nested);
+					AddTypeForwarder(nested, true);
 				}
 			}
 		}
@@ -401,10 +463,12 @@ namespace IKVM.Reflection.Emit
 		private int ExportType(Type type)
 		{
 			ExportedTypeTable.Record rec = new ExportedTypeTable.Record();
-			rec.TypeDefId = type.MetadataToken;
-			rec.TypeName = this.Strings.Add(type.__Name);
-			string ns = type.__Namespace;
-			rec.TypeNamespace = ns == null ? 0 : this.Strings.Add(ns);
+			if (asm.ImageRuntimeVersion == "v2.0.50727")
+			{
+				// HACK we should *not* set the TypeDefId in this case, but 2.0 and 3.5 peverify gives a warning if it is missing (4.5 doesn't)
+				rec.TypeDefId = type.MetadataToken;
+			}
+			SetTypeNameAndTypeNamespace(type.TypeName, out rec.TypeName, out rec.TypeNamespace);
 			if (type.IsNested)
 			{
 				rec.Flags = 0;
@@ -418,6 +482,12 @@ namespace IKVM.Reflection.Emit
 			return 0x27000000 | this.ExportedType.FindOrAddRecord(rec);
 		}
 
+		private void SetTypeNameAndTypeNamespace(TypeName name, out int typeName, out int typeNamespace)
+		{
+			typeName = this.Strings.Add(name.Name);
+			typeNamespace = name.Namespace == null ? 0 : this.Strings.Add(name.Namespace);
+		}
+
 		public void SetCustomAttribute(ConstructorInfo con, byte[] binaryAttribute)
 		{
 			SetCustomAttribute(new CustomAttributeBuilder(con, binaryAttribute));
@@ -430,7 +500,6 @@ namespace IKVM.Reflection.Emit
 
 		internal void SetCustomAttribute(int token, CustomAttributeBuilder customBuilder)
 		{
-			Debug.Assert(!customBuilder.IsPseudoCustomAttribute);
 			CustomAttributeTable.Record rec = new CustomAttributeTable.Record();
 			rec.Parent = token;
 			rec.Type = asm.IsWindowsRuntime ? customBuilder.Constructor.ImportTo(this) : GetConstructorToken(customBuilder.Constructor).Token;
@@ -438,16 +507,23 @@ namespace IKVM.Reflection.Emit
 			this.CustomAttribute.AddRecord(rec);
 		}
 
-		internal void AddDeclarativeSecurity(int token, System.Security.Permissions.SecurityAction securityAction, System.Security.PermissionSet permissionSet)
+		private void AddDeclSecurityRecord(int token, int action, int blob)
 		{
 			DeclSecurityTable.Record rec = new DeclSecurityTable.Record();
-			rec.Action = (short)securityAction;
+			rec.Action = (short)action;
 			rec.Parent = token;
-			// like Ref.Emit, we're using the .NET 1.x xml format
-			rec.PermissionSet = this.Blobs.Add(ByteBuffer.Wrap(System.Text.Encoding.Unicode.GetBytes(permissionSet.ToXml().ToString())));
+			rec.PermissionSet = blob;
 			this.DeclSecurity.AddRecord(rec);
 		}
 
+#if !CORECLR
+		internal void AddDeclarativeSecurity(int token, System.Security.Permissions.SecurityAction securityAction, System.Security.PermissionSet permissionSet)
+		{
+			// like Ref.Emit, we're using the .NET 1.x xml format
+			AddDeclSecurityRecord(token, (int)securityAction, this.Blobs.Add(ByteBuffer.Wrap(System.Text.Encoding.Unicode.GetBytes(permissionSet.ToXml().ToString()))));
+		}
+#endif
+
 		internal void AddDeclarativeSecurity(int token, List<CustomAttributeBuilder> declarativeSecurity)
 		{
 			Dictionary<int, List<CustomAttributeBuilder>> ordered = new Dictionary<int, List<CustomAttributeBuilder>>();
@@ -463,6 +539,11 @@ namespace IKVM.Reflection.Emit
 				{
 					action = (int)cab.GetConstructorArgument(0);
 				}
+				if (cab.IsLegacyDeclSecurity)
+				{
+					AddDeclSecurityRecord(token, action, cab.WriteLegacyDeclSecurityBlob(this));
+					continue;
+				}
 				List<CustomAttributeBuilder> list;
 				if (!ordered.TryGetValue(action, out list))
 				{
@@ -473,22 +554,12 @@ namespace IKVM.Reflection.Emit
 			}
 			foreach (KeyValuePair<int, List<CustomAttributeBuilder>> kv in ordered)
 			{
-				DeclSecurityTable.Record rec = new DeclSecurityTable.Record();
-				rec.Action = (short)kv.Key;
-				rec.Parent = token;
-				rec.PermissionSet = WriteDeclSecurityBlob(kv.Value);
-				this.DeclSecurity.AddRecord(rec);
+				AddDeclSecurityRecord(token, kv.Key, WriteDeclSecurityBlob(kv.Value));
 			}
 		}
 
 		private int WriteDeclSecurityBlob(List<CustomAttributeBuilder> list)
 		{
-			string xml;
-			if (list.Count == 1 && (xml = list[0].GetLegacyDeclSecurity()) != null)
-			{
-				// write .NET 1.1 format
-				return this.Blobs.Add(ByteBuffer.Wrap(System.Text.Encoding.Unicode.GetBytes(xml)));
-			}
 			ByteBuffer namedArgs = new ByteBuffer(100);
 			ByteBuffer bb = new ByteBuffer(list.Count * 100);
 			bb.Write((byte)'.');
@@ -506,21 +577,10 @@ namespace IKVM.Reflection.Emit
 
 		public void DefineManifestResource(string name, Stream stream, ResourceAttributes attribute)
 		{
-			manifestResources.Align(8);
-			ManifestResourceTable.Record rec = new ManifestResourceTable.Record();
-			rec.Offset = manifestResources.Position;
-			rec.Flags = (int)attribute;
-			rec.Name = this.Strings.Add(name);
-			rec.Implementation = 0;
-			this.ManifestResource.AddRecord(rec);
-			manifestResources.Write(0);	// placeholder for the length
-			manifestResources.Write(stream);
-			int savePosition = manifestResources.Position;
-			manifestResources.Position = rec.Offset;
-			manifestResources.Write(savePosition - (manifestResources.Position + 4));
-			manifestResources.Position = savePosition;
+			resourceWriters.Add(new ResourceWriterRecord(name, stream, attribute));
 		}
 
+#if !CORECLR
 		public IResourceWriter DefineResource(string name, string description)
 		{
 			return DefineResource(name, description, ResourceAttributes.Public);
@@ -530,27 +590,61 @@ namespace IKVM.Reflection.Emit
 		{
 			// FXBUG we ignore the description, because there is no such thing
 
-			if (resourceWriters == null)
-			{
-				resourceWriters = new List<ResourceWriterRecord>();
-			}
 			MemoryStream mem = new MemoryStream();
 			ResourceWriter rw = new ResourceWriter(mem);
 			resourceWriters.Add(new ResourceWriterRecord(name, rw, mem, attribute));
 			return rw;
 		}
+#endif
 
 		internal void EmitResources()
 		{
-			if (resourceWriters != null)
+			int offset = 0;
+			foreach (ResourceWriterRecord rwr in resourceWriters)
 			{
-				foreach (ResourceWriterRecord rwr in resourceWriters)
-				{
-					rwr.Emit(this);
-				}
+				// resources must be 8-byte aligned
+				offset = (offset + 7) & ~7;
+				rwr.Emit(this, offset);
+				offset += rwr.GetLength();
 			}
 		}
 
+		internal void WriteResources(MetadataWriter mw)
+		{
+			int offset = 0;
+			foreach (ResourceWriterRecord rwr in resourceWriters)
+			{
+				// resources must be 8-byte aligned
+				int alignment = ((offset + 7) & ~7) - offset;
+				for (int i = 0; i < alignment; i++)
+				{
+					mw.Write((byte)0);
+				}
+				rwr.Write(mw);
+				offset += rwr.GetLength() + alignment;
+			}
+		}
+
+		internal void CloseResources()
+		{
+			foreach (ResourceWriterRecord rwr in resourceWriters)
+			{
+				rwr.Close();
+			}
+		}
+
+		internal int GetManifestResourcesLength()
+		{
+			int length = 0;
+			foreach (ResourceWriterRecord rwr in resourceWriters)
+			{
+				// resources must be 8-byte aligned
+				length = (length + 7) & ~7;
+				length += rwr.GetLength();
+			}
+			return length;
+		}
+
 		public override Assembly Assembly
 		{
 			get { return asm; }
@@ -560,7 +654,7 @@ namespace IKVM.Reflection.Emit
 		{
 			foreach (Type type in types)
 			{
-				if (type.__Namespace == name.Namespace && type.__Name == name.Name)
+				if (type.TypeName == name)
 				{
 					return type;
 				}
@@ -572,7 +666,7 @@ namespace IKVM.Reflection.Emit
 		{
 			foreach (Type type in types)
 			{
-				if (new TypeName(type.__Namespace, type.__Name).ToLowerInvariant() == lowerCaseName)
+				if (type.TypeName.ToLowerInvariant() == lowerCaseName)
 				{
 					return type;
 				}
@@ -591,10 +685,12 @@ namespace IKVM.Reflection.Emit
 			}
 		}
 
+#if !NO_SYMBOL_WRITER
 		public ISymbolDocumentWriter DefineDocument(string url, Guid language, Guid languageVendor, Guid documentType)
 		{
 			return symbolWriter.DefineDocument(url, language, languageVendor, documentType);
 		}
+#endif
 
 		public int __GetAssemblyToken(Assembly assembly)
 		{
@@ -816,9 +912,7 @@ namespace IKVM.Reflection.Emit
 					{
 						rec.ResolutionScope = ImportAssemblyRef(type.Assembly);
 					}
-					rec.TypeName = this.Strings.Add(type.__Name);
-					string ns = type.__Namespace;
-					rec.TypeNameSpace = ns == null ? 0 : this.Strings.Add(ns);
+					SetTypeNameAndTypeNamespace(type.TypeName, out rec.TypeName, out rec.TypeNamespace);
 					token = 0x01000000 | this.TypeRef.AddRecord(rec);
 				}
 				typeTokens.Add(type, token);
@@ -920,7 +1014,7 @@ namespace IKVM.Reflection.Emit
 			resolvedTokens[index] = realToken;
 		}
 
-		internal bool IsPseudoToken(int token)
+		internal static bool IsPseudoToken(int token)
 		{
 			return token < 0;
 		}
@@ -937,15 +1031,19 @@ namespace IKVM.Reflection.Emit
 			{
 				int type;
 				int size;
-				if (imageFileMachine == ImageFileMachine.I386)
+				switch (imageFileMachine)
 				{
-					type = 0x05;
-					size = 4;
-				}
-				else
-				{
-					type = 0x06;
-					size = 8;
+					case ImageFileMachine.I386:
+					case ImageFileMachine.ARM:
+						type = 0x05;
+						size = 4;
+						break;
+					case ImageFileMachine.AMD64:
+						type = 0x06;
+						size = 8;
+						break;
+					default:
+						throw new NotSupportedException();
 				}
 				List<MethodBuilder> methods = new List<MethodBuilder>();
 				for (int i = 0; i < unmanagedExports.Count; i++)
@@ -1035,7 +1133,7 @@ namespace IKVM.Reflection.Emit
 			}
 		}
 
-		internal void WriteMetadata(MetadataWriter mw)
+		internal void WriteMetadata(MetadataWriter mw, out int guidHeapOffset)
 		{
 			mw.Write(0x424A5342);			// Signature ("BSJB")
 			mw.Write((ushort)1);			// MajorVersion
@@ -1088,6 +1186,7 @@ namespace IKVM.Reflection.Emit
 			Tables.Write(mw);
 			Strings.Write(mw);
 			UserStrings.Write(mw);
+			guidHeapOffset = mw.Position;
 			Guids.Write(mw);
 			if (!Blobs.IsEmpty)
 			{
@@ -1123,9 +1222,7 @@ namespace IKVM.Reflection.Emit
 					rec.Flags = (int)type.Attributes;
 					// LAMESPEC ECMA says that TypeDefId is a row index, but it should be a token
 					rec.TypeDefId = type.MetadataToken;
-					rec.TypeName = this.Strings.Add(type.__Name);
-					string ns = type.__Namespace;
-					rec.TypeNamespace = ns == null ? 0 : this.Strings.Add(ns);
+					SetTypeNameAndTypeNamespace(type.TypeName, out rec.TypeName, out rec.TypeNamespace);
 					if (type.IsNested)
 					{
 						rec.Implementation = declaringTypes[type.DeclaringType];
@@ -1309,16 +1406,52 @@ namespace IKVM.Reflection.Emit
 			get { return fileName; }
 		}
 
+		internal Guid GetModuleVersionIdOrEmpty()
+		{
+			return mvid;
+		}
+
 		public override Guid ModuleVersionId
 		{
-			get { return mvid; }
+			get
+			{
+				if (mvid == Guid.Empty && universe.Deterministic)
+				{
+					// if a deterministic GUID is used, it can't be queried before the assembly has been written
+					throw new InvalidOperationException();
+				}
+				return mvid;
+			}
 		}
 
 		public void __SetModuleVersionId(Guid guid)
 		{
+			if (guid == Guid.Empty && universe.Deterministic)
+			{
+				// if you want to use Guid.Empty, don't set UniverseOptions.DeterministicOutput
+				throw new ArgumentOutOfRangeException();
+			}
 			mvid = guid;
 		}
 
+		internal uint GetTimeDateStamp()
+		{
+			return timestamp;
+		}
+
+		public DateTime __PEHeaderTimeDateStamp
+		{
+			get { return new DateTime(1970, 1, 1, 0, 0, 0, DateTimeKind.Utc).AddSeconds(timestamp); }
+			set
+			{
+				if (value < new DateTime(1970, 1, 1, 0, 0, 0, DateTimeKind.Utc) || value > new DateTime(2106, 2, 7, 6, 28, 15, DateTimeKind.Utc))
+				{
+					throw new ArgumentOutOfRangeException();
+				}
+				timestamp = (uint)(value - new DateTime(1970, 1, 1, 0, 0, 0, DateTimeKind.Utc)).TotalSeconds;
+			}
+		}
+
 		public override Type[] __ResolveOptionalParameterTypes(int metadataToken, Type[] genericTypeArguments, Type[] genericMethodArguments, out CustomModifiers[] customModifiers)
 		{
 			throw new NotImplementedException();
@@ -1329,10 +1462,12 @@ namespace IKVM.Reflection.Emit
 			get { return moduleName; }
 		}
 
+#if !NO_SYMBOL_WRITER
 		public ISymbolWriter GetSymWriter()
 		{
 			return symbolWriter;
 		}
+#endif
 
 		public void DefineUnmanagedResource(string resourceFileName)
 		{
@@ -1354,10 +1489,12 @@ namespace IKVM.Reflection.Emit
 			{
 				token = -token | 0x06000000;
 			}
+#if !NO_SYMBOL_WRITER
 			if (symbolWriter != null)
 			{
 				symbolWriter.SetUserEntryPoint(new SymbolToken(token));
 			}
+#endif
 		}
 
 		public StringToken GetStringConstant(string str)
@@ -1467,8 +1604,7 @@ namespace IKVM.Reflection.Emit
 		{
 			TypeRefTable.Record rec = new TypeRefTable.Record();
 			rec.ResolutionScope = resolutionScope;
-			rec.TypeName = this.Strings.Add(name);
-			rec.TypeNameSpace = ns == null ? 0 : this.Strings.Add(ns);
+			SetTypeNameAndTypeNamespace(new TypeName(ns, name), out rec.TypeName, out rec.TypeNamespace);
 			return 0x01000000 | this.TypeRef.AddRecord(rec);
 		}
 
@@ -1529,6 +1665,7 @@ namespace IKVM.Reflection.Emit
 			FillAssemblyRefTable();
 			EmitResources();
 			ModuleWriter.WriteModule(null, null, this, PEFileKinds.Dll, portableExecutableKind, imageFileMachine, unmanagedResources, 0, streamOrNull);
+			CloseResources();
 		}
 
 		public void __AddAssemblyReference(AssemblyName assemblyName)
