@@ -23,7 +23,9 @@
 */
 using System;
 using System.Runtime.InteropServices;
+#if !NO_SYMBOL_WRITER
 using System.Diagnostics.SymbolStore;
+#endif
 using IKVM.Reflection.Emit;
 
 namespace IKVM.Reflection.Impl
@@ -41,25 +43,39 @@ namespace IKVM.Reflection.Impl
 		public uint PointerToRawData;
 	}
 
+#if NO_SYMBOL_WRITER
+	struct SymbolToken
+	{
+		internal SymbolToken(int value) { }
+	}
+
+	interface ISymbolWriterImpl
+	{
+		byte[] GetDebugInfo(ref IMAGE_DEBUG_DIRECTORY idd);
+		void RemapToken(int oldToken, int newToken);
+		void DefineLocalVariable2(string name, FieldAttributes attributes, int signature, int addrKind, int addr1, int addr2, int addr3, int startOffset, int endOffset);
+		void OpenMethod(SymbolToken symbolToken, MethodBase mb);
+		bool IsDeterministic { get; }
+	}
+#else
 	interface ISymbolWriterImpl : ISymbolWriter
 	{
 		byte[] GetDebugInfo(ref IMAGE_DEBUG_DIRECTORY idd);
 		void RemapToken(int oldToken, int newToken);
 		void DefineLocalVariable2(string name, FieldAttributes attributes, int signature, SymAddressKind addrKind, int addr1, int addr2, int addr3, int startOffset, int endOffset);
+		void OpenMethod(SymbolToken symbolToken, MethodBase mb);
+		bool IsDeterministic { get; }
 	}
+#endif
 
 	static class SymbolSupport
 	{
-#if !NO_SYMBOL_WRITER
-		private static readonly bool runningOnMono = System.Type.GetType("Mono.Runtime") != null;
-#endif
-
 		internal static ISymbolWriterImpl CreateSymbolWriterFor(ModuleBuilder moduleBuilder)
 		{
 #if NO_SYMBOL_WRITER
 			throw new NotSupportedException("IKVM.Reflection compiled with NO_SYMBOL_WRITER does not support writing debugging symbols.");
 #else
-			if (runningOnMono)
+			if (Universe.MonoRuntime)
 			{
 #if MONO
 				return new MdbWriter(moduleBuilder);
