@@ -211,7 +211,7 @@ enum ByteCode : byte
 	__invokespecial = 183,
 	__invokestatic = 184,
 	__invokeinterface = 185,
-	__xxxunusedxxx = 186,
+	__invokedynamic = 186,
 	__new = 187,
 	__newarray = 188,
 	__anewarray = 189,
@@ -366,6 +366,7 @@ enum NormalizedByteCode : byte
 	__invokespecial = 183,
 	__invokestatic = 184,
 	__invokeinterface = 185,
+	__invokedynamic = 186,
 	__new = 187,
 	__newarray = 188,
 	__anewarray = 189,
@@ -379,6 +380,12 @@ enum NormalizedByteCode : byte
 	__ifnull = 198,
 	__ifnonnull = 199,
 	// This is where the pseudo-bytecodes start
+	__privileged_invokestatic = 235,		// the privileged bytecodes are used for accessing host class members
+	__privileged_invokevirtual = 237,
+	__privileged_invokespecial = 238,
+	__ldc_nothrow = 239,
+	__methodhandle_invoke = 240,
+	__methodhandle_link = 241,
 	__goto_finally = 242,
 	__intrinsic_gettype = 243,
 	__athrow_no_unmap = 244,
@@ -575,8 +582,11 @@ struct ByteCodeMetaData
 			case NormalizedByteCode.__dynamic_putfield:
 			case NormalizedByteCode.__clone_array:
 			case NormalizedByteCode.__static_error:
+			case NormalizedByteCode.__methodhandle_invoke:
+			case NormalizedByteCode.__methodhandle_link:
 				return true;
 			case NormalizedByteCode.__iconst:
+			case NormalizedByteCode.__ldc_nothrow:
 				return false;
 			default:
 				return (data[(int)bc].flags & ByteCodeFlags.CannotThrow) == 0;
@@ -617,9 +627,9 @@ struct ByteCodeMetaData
 		new ByteCodeMetaData(ByteCode.__dconst_1, ByteCodeMode.Simple, ByteCodeModeWide.Unused, true);
 		new ByteCodeMetaData(ByteCode.__bipush, NormalizedByteCode.__iconst, ByteCodeMode.Immediate_1, ByteCodeModeWide.Unused, true);
 		new ByteCodeMetaData(ByteCode.__sipush, NormalizedByteCode.__iconst, ByteCodeMode.Immediate_2, ByteCodeModeWide.Unused, true);
-		new ByteCodeMetaData(ByteCode.__ldc, ByteCodeMode.Constant_1, ByteCodeModeWide.Unused, true);
-		new ByteCodeMetaData(ByteCode.__ldc_w, NormalizedByteCode.__ldc, ByteCodeMode.Constant_2, ByteCodeModeWide.Unused, true);
-		new ByteCodeMetaData(ByteCode.__ldc2_w, NormalizedByteCode.__ldc, ByteCodeMode.Constant_2, ByteCodeModeWide.Unused, true);
+		new ByteCodeMetaData(ByteCode.__ldc, ByteCodeMode.Constant_1, ByteCodeModeWide.Unused, false);
+		new ByteCodeMetaData(ByteCode.__ldc_w, NormalizedByteCode.__ldc, ByteCodeMode.Constant_2, ByteCodeModeWide.Unused, false);
+		new ByteCodeMetaData(ByteCode.__ldc2_w, NormalizedByteCode.__ldc, ByteCodeMode.Constant_2, ByteCodeModeWide.Unused, false);
 		new ByteCodeMetaData(ByteCode.__iload, ByteCodeMode.Local_1, ByteCodeModeWide.Local_2, true);
 		new ByteCodeMetaData(ByteCode.__lload, ByteCodeMode.Local_1, ByteCodeModeWide.Local_2, true);
 		new ByteCodeMetaData(ByteCode.__fload, ByteCodeMode.Local_1, ByteCodeModeWide.Local_2, true);
@@ -785,7 +795,7 @@ struct ByteCodeMetaData
 		new ByteCodeMetaData(ByteCode.__invokespecial, ByteCodeMode.Constant_2, ByteCodeModeWide.Unused, false);
 		new ByteCodeMetaData(ByteCode.__invokestatic, ByteCodeMode.Constant_2, ByteCodeModeWide.Unused, false);
 		new ByteCodeMetaData(ByteCode.__invokeinterface, ByteCodeMode.Constant_2_1_1, ByteCodeModeWide.Unused, false);
-		new ByteCodeMetaData(ByteCode.__xxxunusedxxx, NormalizedByteCode.__nop, ByteCodeMode.Unused, ByteCodeModeWide.Unused, true);
+		new ByteCodeMetaData(ByteCode.__invokedynamic, ByteCodeMode.Constant_2_1_1, ByteCodeModeWide.Unused, false);
 		new ByteCodeMetaData(ByteCode.__new, ByteCodeMode.Constant_2, ByteCodeModeWide.Unused, false);
 		new ByteCodeMetaData(ByteCode.__newarray, ByteCodeMode.Immediate_1, ByteCodeModeWide.Unused, false);
 		new ByteCodeMetaData(ByteCode.__anewarray, ByteCodeMode.Constant_2, ByteCodeModeWide.Unused, false);
