@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2002-2012 Jeroen Frijters
+  Copyright (C) 2002-2015 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -146,153 +146,92 @@ namespace IKVM.Runtime
 			return obj;
 		}
 
-#if !FIRST_PASS
 		[DebuggerStepThroughAttribute]
-		public static object DynamicMultianewarray(RuntimeTypeHandle type, string clazz, int[] lengths)
+		public static object DynamicMultianewarray(int[] lengths, java.lang.Class clazz)
 		{
+#if FIRST_PASS
+			return null;
+#else
 			Profiler.Count("DynamicMultianewarray");
-			TypeWrapper wrapper = LoadTypeWrapper(type, clazz);
-			return multianewarray(wrapper.TypeAsArrayType.TypeHandle, lengths);
+			TypeWrapper wrapper = TypeWrapper.FromClass(clazz);
+			object obj = multianewarray(wrapper.TypeAsArrayType.TypeHandle, lengths);
+			if (wrapper.IsGhostArray)
+			{
+				GhostTag.SetTag(obj, wrapper);
+			}
+			return obj;
+#endif
 		}
 
 		[DebuggerStepThroughAttribute]
-		public static object DynamicNewarray(int length, RuntimeTypeHandle type, string clazz)
+		public static object DynamicNewarray(int length, java.lang.Class clazz)
 		{
+#if FIRST_PASS
+			return null;
+#else
 			Profiler.Count("DynamicNewarray");
 			if(length < 0)
 			{
 				throw new java.lang.NegativeArraySizeException();
 			}
-			TypeWrapper wrapper = LoadTypeWrapper(type, clazz);
-			return Array.CreateInstance(wrapper.TypeAsArrayType, length);
+			TypeWrapper wrapper = TypeWrapper.FromClass(clazz);
+			Array obj = Array.CreateInstance(wrapper.TypeAsArrayType, length);
+			if (wrapper.IsGhost || wrapper.IsGhostArray)
+			{
+				GhostTag.SetTag(obj, wrapper.MakeArrayType(1));
+			}
+			return obj;
+#endif
 		}
 
 		[DebuggerStepThroughAttribute]
-		public static void DynamicAastore(object arrayref, int index, object val, RuntimeTypeHandle type, string clazz)
+		public static void DynamicAastore(object arrayref, int index, object val)
 		{
+#if !FIRST_PASS
 			Profiler.Count("DynamicAastore");
-			// TODO do we need to load the type here?
 			((Array)arrayref).SetValue(val, index);
+#endif
 		}
 
 		[DebuggerStepThroughAttribute]
-		public static object DynamicAaload(object arrayref, int index, RuntimeTypeHandle type, string clazz)
+		public static object DynamicAaload(object arrayref, int index)
 		{
+#if FIRST_PASS
+			return null;
+#else
 			Profiler.Count("DynamicAaload");
-			// TODO do we need to load the type here?
 			return ((Array)arrayref).GetValue(index);
-		}
-
-		private static FieldWrapper GetFieldWrapper(object thisObj, RuntimeTypeHandle type, string clazz, string name, string sig, bool isStatic)
-		{
-			TypeWrapper caller = ClassLoaderWrapper.GetWrapperFromType(Type.GetTypeFromHandle(type));
-			TypeWrapper wrapper = LoadTypeWrapper(type, clazz);
-			FieldWrapper field = wrapper.GetFieldWrapper(name, sig);
-			if(field == null)
-			{
-				throw new java.lang.NoSuchFieldError(clazz + "." + name);
-			}
-			// TODO check loader constraints
-			if(field.IsStatic != isStatic)
-			{
-				throw new java.lang.IncompatibleClassChangeError(clazz + "." + name);
-			}
-			TypeWrapper objType = null;
-			if(thisObj != null)
-			{
-				objType = ClassLoaderWrapper.GetWrapperFromType(thisObj.GetType());
-			}
-			if(field.IsAccessibleFrom(wrapper, caller, objType))
-			{
-				return field;
-			}
-			throw new java.lang.IllegalAccessError(field.DeclaringType.Name + "." + name);
-		}
-
-		[DebuggerStepThroughAttribute]
-		public static object DynamicGetfield(object obj, string name, string sig, RuntimeTypeHandle type, string clazz, ikvm.@internal.CallerID callerID)
-		{
-			Profiler.Count("DynamicGetfield");
-			FieldWrapper fw = GetFieldWrapper(obj, type, clazz, name, sig, false);
-			java.lang.reflect.Field field = (java.lang.reflect.Field)fw.ToField(false);
-			object val = field.get(obj, callerID);
-			if(fw.FieldTypeWrapper.IsPrimitive)
-			{
-				val = JVM.Unbox(val);
-			}
-			return val;
-		}
-
-		[DebuggerStepThroughAttribute]
-		public static object DynamicGetstatic(string name, string sig, RuntimeTypeHandle type, string clazz, ikvm.@internal.CallerID callerID)
-		{
-			Profiler.Count("DynamicGetstatic");
-			FieldWrapper fw = GetFieldWrapper(null, type, clazz, name, sig, true);
-			java.lang.reflect.Field field = (java.lang.reflect.Field)fw.ToField(false);
-			object val = field.get(null, callerID);
-			if(fw.FieldTypeWrapper.IsPrimitive)
-			{
-				val = JVM.Unbox(val);
-			}
-			return val;
-		}
-
-		[DebuggerStepThroughAttribute]
-		public static void DynamicPutfield(object obj, object val, string name, string sig, RuntimeTypeHandle type, string clazz, ikvm.@internal.CallerID callerID)
-		{
-			Profiler.Count("DynamicPutfield");
-			FieldWrapper fw = GetFieldWrapper(obj, type, clazz, name, sig, false);
-			if(fw.IsFinal)
-			{
-				throw new java.lang.IllegalAccessError("Field " + fw.DeclaringType.Name + "." + fw.Name + " is final");
-			}
-			java.lang.reflect.Field field = (java.lang.reflect.Field)fw.ToField(false);
-			if(fw.FieldTypeWrapper.IsPrimitive)
-			{
-				val = JVM.Box(val);
-			}
-			field.set(obj, val, callerID);
-		}
-
-		[DebuggerStepThroughAttribute]
-		public static void DynamicPutstatic(object val, string name, string sig, RuntimeTypeHandle type, string clazz, ikvm.@internal.CallerID callerID)
-		{
-			Profiler.Count("DynamicPutstatic");
-			FieldWrapper fw = GetFieldWrapper(null, type, clazz, name, sig, true);
-			if(fw.IsFinal)
-			{
-				throw new java.lang.IllegalAccessError("Field " + fw.DeclaringType.Name + "." + fw.Name + " is final");
-			}
-			java.lang.reflect.Field field = (java.lang.reflect.Field)fw.ToField(false);
-			if(fw.FieldTypeWrapper.IsPrimitive)
-			{
-				val = JVM.Box(val);
-			}
-			field.set(null, val, callerID);
+#endif
 		}
 
 		// the sole purpose of this method is to check whether the clazz can be instantiated (but not to actually do it)
 		[DebuggerStepThroughAttribute]
-		public static void DynamicNewCheckOnly(RuntimeTypeHandle type, string clazz)
+		public static void DynamicNewCheckOnly(java.lang.Class clazz)
 		{
+#if !FIRST_PASS
 			Profiler.Count("DynamicNewCheckOnly");
-			TypeWrapper wrapper = LoadTypeWrapper(type, clazz);
+			TypeWrapper wrapper = TypeWrapper.FromClass(clazz);
 			if(wrapper.IsAbstract)
 			{
-				throw new java.lang.InstantiationError(clazz);
+				throw new java.lang.InstantiationError(wrapper.Name);
 			}
 			wrapper.RunClassInit();
+#endif
 		}
 
-		private static TypeWrapper LoadTypeWrapper(RuntimeTypeHandle type, string clazz)
+		private static TypeWrapper LoadTypeWrapper(string clazz, ikvm.@internal.CallerID callerId)
 		{
+#if FIRST_PASS
+			return null;
+#else
 			try
 			{
-				TypeWrapper context = ClassLoaderWrapper.GetWrapperFromType(Type.GetTypeFromHandle(type));
-				TypeWrapper wrapper = context.GetClassLoader().LoadClassByDottedNameFast(clazz);
-				if(wrapper == null)
+				TypeWrapper context = TypeWrapper.FromClass(callerId.getCallerClass());
+				TypeWrapper wrapper = ClassLoaderWrapper.FromCallerID(callerId).LoadClassByDottedName(clazz);
+				java.lang.ClassLoader loader = callerId.getCallerClassLoader();
+				if(loader != null)
 				{
-					throw new java.lang.NoClassDefFoundError(clazz);
+					loader.checkPackageAccess(wrapper.ClassObject, callerId.getCallerClass().pd);
 				}
 				if(!wrapper.IsAccessibleFrom(context))
 				{
@@ -305,132 +244,199 @@ namespace IKVM.Runtime
 			{
 				throw x.ToJava();
 			}
+#endif
 		}
 
 		[DebuggerStepThroughAttribute]
-		public static object DynamicClassLiteral(RuntimeTypeHandle type, string clazz)
+		public static java.lang.Class DynamicClassLiteral(string clazz, ikvm.@internal.CallerID callerId)
 		{
+#if FIRST_PASS
+			return null;
+#else
 			Profiler.Count("DynamicClassLiteral");
-			return LoadTypeWrapper(type, clazz).ClassObject;
+			return LoadTypeWrapper(clazz, callerId).ClassObject;
+#endif
 		}
 
 		[DebuggerStepThroughAttribute]
-		public static object DynamicCast(object obj, RuntimeTypeHandle type, string clazz)
+		public static object DynamicCast(object obj, java.lang.Class clazz)
 		{
+#if FIRST_PASS
+			return null;
+#else
+			Debug.Assert(obj != null);
 			Profiler.Count("DynamicCast");
-			// NOTE it's important that we don't try to load the class if obj == null
-			// (to be compatible with Sun)
-			if(obj != null && !DynamicInstanceOf(obj, type, clazz))
+			if (!DynamicInstanceOf(obj, clazz))
 			{
-				throw new java.lang.ClassCastException(NativeCode.ikvm.runtime.Util.GetTypeWrapperFromObject(obj).Name);
+				throw new java.lang.ClassCastException(NativeCode.ikvm.runtime.Util.GetTypeWrapperFromObject(obj).Name + " cannot be cast to " + clazz.getName());
 			}
 			return obj;
+#endif
 		}
 
 		[DebuggerStepThroughAttribute]
-		public static bool DynamicInstanceOf(object obj, RuntimeTypeHandle type, string clazz)
+		public static bool DynamicInstanceOf(object obj, java.lang.Class clazz)
 		{
+#if FIRST_PASS
+			return false;
+#else
+			Debug.Assert(obj != null);
 			Profiler.Count("DynamicInstanceOf");
-			// NOTE it's important that we don't try to load the class if obj == null
-			// (to be compatible with Sun)
-			if(obj == null)
+			TypeWrapper tw = TypeWrapper.FromClass(clazz);
+			// we have to mimick the bytecode behavior, which allows these .NET-isms to show through
+			if (tw.TypeAsBaseType == typeof(Array))
 			{
-				return false;
+				return obj is Array;
 			}
-			TypeWrapper wrapper = LoadTypeWrapper(type, clazz);
-			return wrapper.IsInstance(obj);
+			if (tw.TypeAsBaseType == typeof(string))
+			{
+				return obj is string;
+			}
+			if (tw.TypeAsBaseType == typeof(IComparable))
+			{
+				return obj is IComparable;
+			}
+			return tw.IsInstance(obj);
+#endif
 		}
 
-		private static MethodWrapper GetMethodWrapper(object thisObj, RuntimeTypeHandle type, string clazz, string name, string sig, bool isStatic)
+		[DebuggerStepThrough]
+		public static java.lang.invoke.MethodType DynamicLoadMethodType(ref java.lang.invoke.MethodType cache, string sig, ikvm.@internal.CallerID callerID)
 		{
-			TypeWrapper caller = ClassLoaderWrapper.GetWrapperFromType(Type.GetTypeFromHandle(type));
-			TypeWrapper wrapper = LoadTypeWrapper(type, clazz);
-			MethodWrapper mw = wrapper.GetMethodWrapper(name, sig, false);
-			if(mw == null)
+			if (cache == null)
 			{
-				throw new java.lang.NoSuchMethodError(clazz + "." + name + sig);
+				DynamicLoadMethodTypeImpl(ref cache, sig, callerID);
 			}
-			// TODO check loader constraints
-			if(mw.IsStatic != isStatic)
-			{
-				throw new java.lang.IncompatibleClassChangeError(clazz + "." + name);
-			}
-			TypeWrapper objType = null;
-			if(thisObj != null)
-			{
-				objType = ClassLoaderWrapper.GetWrapperFromType(thisObj.GetType());
-			}
-			if(mw.IsAccessibleFrom(wrapper, caller, objType))
-			{
-				return mw;
-			}
-			throw new java.lang.IllegalAccessError(clazz + "." + name + sig);
+			return cache;
 		}
 
-		[DebuggerStepThroughAttribute]
-		public static object DynamicInvokeSpecialNew(RuntimeTypeHandle type, string clazz, string name, string sig, object[] args, ikvm.@internal.CallerID callerID)
+		private static void DynamicLoadMethodTypeImpl(ref java.lang.invoke.MethodType cache, string sig, ikvm.@internal.CallerID callerID)
 		{
-			Profiler.Count("DynamicInvokeSpecialNew");
-			MethodWrapper mw = GetMethodWrapper(null, type, clazz, name, sig, false);
-			java.lang.reflect.Constructor cons = (java.lang.reflect.Constructor)mw.ToMethodOrConstructor(false);
-			return cons.newInstance(BoxArgs(mw, args), callerID);
-		}
-
-		[DebuggerStepThroughAttribute]
-		public static object DynamicInvokestatic(RuntimeTypeHandle type, string clazz, string name, string sig, object[] args, ikvm.@internal.CallerID callerID)
-		{
-			Profiler.Count("DynamicInvokestatic");
-			MethodWrapper mw = GetMethodWrapper(null, type, clazz, name, sig, true);
-			java.lang.reflect.Method m = (java.lang.reflect.Method)mw.ToMethodOrConstructor(false);
-			object val = m.invoke(null, BoxArgs(mw, args), callerID);
-			if (mw.ReturnType.IsPrimitive && mw.ReturnType != PrimitiveTypeWrapper.VOID)
+#if !FIRST_PASS
+			try
 			{
-				val = JVM.Unbox(val);
-			}
-			return val;
-		}
-
-		[DebuggerStepThroughAttribute]
-		public static object DynamicInvokevirtual(object obj, RuntimeTypeHandle type, string clazz, string name, string sig, object[] args, ikvm.@internal.CallerID callerID)
-		{
-			Profiler.Count("DynamicInvokevirtual");
-			MethodWrapper mw = GetMethodWrapper(obj, type, clazz, name, sig, false);
-			java.lang.reflect.Method m = (java.lang.reflect.Method)mw.ToMethodOrConstructor(false);
-			object val = m.invoke(obj, BoxArgs(mw, args), callerID);
-			if (mw.ReturnType.IsPrimitive && mw.ReturnType != PrimitiveTypeWrapper.VOID)
-			{
-				val = JVM.Unbox(val);
-			}
-			return val;
-		}
-
-		private static object[] BoxArgs(MethodWrapper mw, object[] args)
-		{
-			TypeWrapper[] paramTypes = mw.GetParameters();
-			for (int i = 0; i < paramTypes.Length; i++)
-			{
-				if (paramTypes[i].IsPrimitive)
+				ClassLoaderWrapper loader = ClassLoaderWrapper.FromCallerID(callerID);
+				TypeWrapper[] args = loader.ArgTypeWrapperListFromSig(sig, LoadMode.LoadOrThrow);
+				java.lang.Class[] ptypes = new java.lang.Class[args.Length];
+				for (int i = 0; i < ptypes.Length; i++)
 				{
-					args[i] = JVM.Box(args[i]);
+					ptypes[i] = args[i].ClassObject;
+				}
+				Interlocked.CompareExchange(ref cache, java.lang.invoke.MethodType.methodType(loader.RetTypeWrapperFromSig(sig, LoadMode.LoadOrThrow).ClassObject, ptypes), null);
+			}
+			catch (RetargetableJavaException x)
+			{
+				throw x.ToJava();
+			}
+#endif
+		}
+
+		[DebuggerStepThrough]
+		public static java.lang.invoke.MethodHandle DynamicLoadMethodHandle(ref java.lang.invoke.MethodHandle cache, int kind, string clazz, string name, string sig, ikvm.@internal.CallerID callerID)
+		{
+			if (cache == null)
+			{
+				Interlocked.CompareExchange(ref cache, DynamicLoadMethodHandleImpl(kind, clazz, name, sig, callerID), null);
+			}
+			return cache;
+		}
+
+		private static java.lang.invoke.MethodHandle DynamicLoadMethodHandleImpl(int kind, string clazz, string name, string sig, ikvm.@internal.CallerID callerID)
+		{
+#if FIRST_PASS
+			return null;
+#else
+			java.lang.Class refc = LoadTypeWrapper(clazz, callerID).ClassObject;
+			try
+			{
+				switch ((ClassFile.RefKind)kind)
+				{
+					case ClassFile.RefKind.getStatic:
+					case ClassFile.RefKind.putStatic:
+					case ClassFile.RefKind.getField:
+					case ClassFile.RefKind.putField:
+						java.lang.Class type = ClassLoaderWrapper.FromCallerID(callerID).FieldTypeWrapperFromSig(sig, LoadMode.LoadOrThrow).ClassObject;
+						return java.lang.invoke.MethodHandleNatives.linkMethodHandleConstant(callerID.getCallerClass(), kind, refc, name, type);
+					default:
+						java.lang.invoke.MethodType mt = null;
+						DynamicLoadMethodType(ref mt, sig, callerID);
+						// HACK linkMethodHandleConstant is broken for MethodHandle.invoke[Exact]
+						if (kind == (int)ClassFile.RefKind.invokeVirtual && refc == CoreClasses.java.lang.invoke.MethodHandle.Wrapper.ClassObject)
+						{
+							switch (name)
+							{
+								case "invokeExact":
+									return java.lang.invoke.MethodHandles.exactInvoker(mt);
+								case "invoke":
+									return java.lang.invoke.MethodHandles.invoker(mt);
+							}
+						}
+						return java.lang.invoke.MethodHandleNatives.linkMethodHandleConstant(callerID.getCallerClass(), kind, refc, name, mt);
 				}
 			}
-			return args;
+			catch (RetargetableJavaException x)
+			{
+				throw x.ToJava();
+			}
+#endif
 		}
 
-		[DebuggerStepThroughAttribute]
-		public static Type DynamicGetTypeAsExceptionType(RuntimeTypeHandle type, string clazz)
+		[DebuggerStepThrough]
+		public static T DynamicBinderMemberLookup<T>(int kind, string clazz, string name, string sig, ikvm.@internal.CallerID callerID)
+			where T : class /* delegate */
 		{
-			Profiler.Count("DynamicGetTypeAsExceptionType");
-			return LoadTypeWrapper(type, clazz).TypeAsExceptionType;
+#if FIRST_PASS
+			return null;
+#else
+			try
+			{
+				java.lang.invoke.MethodHandle mh = DynamicLoadMethodHandleImpl(kind, clazz, name, sig, callerID);
+				return GetDelegateForInvokeExact<T>(java.lang.invoke.MethodHandles.explicitCastArguments(mh, MethodHandleUtil.GetDelegateMethodType(typeof(T))));
+			}
+			catch (java.lang.IncompatibleClassChangeError x)
+			{
+				if (x.getCause() is java.lang.NoSuchMethodException)
+				{
+					throw new java.lang.NoSuchMethodError(x.getCause().Message);
+				}
+				if (x.getCause() is java.lang.NoSuchFieldException)
+				{
+					throw new java.lang.NoSuchFieldError(x.getCause().Message);
+				}
+				if (x.getCause() is java.lang.IllegalAccessException)
+				{
+					throw new java.lang.IllegalAccessError(x.getCause().Message);
+				}
+				throw;
+			}
+#endif
 		}
 
 		[DebuggerStepThrough]
 		public static Delegate DynamicCreateDelegate(object obj, Type delegateType, string name, string sig)
 		{
+#if FIRST_PASS
+			return null;
+#else
 			TypeWrapper tw = TypeWrapper.FromClass(ikvm.runtime.Util.getClassFromObject(obj));
 			MethodWrapper mw = tw.GetMethodWrapper(name, sig, true);
 			if (mw == null || mw.IsStatic || !mw.IsPublic)
 			{
+#if NO_REF_EMIT
+				java.lang.invoke.MethodType methodType = MethodHandleUtil.GetDelegateMethodType(delegateType);
+				if (methodType.parameterCount() > MethodHandleUtil.MaxArity)
+				{
+					throw new NotImplementedException();
+				}
+				java.lang.invoke.MethodHandle exception = java.lang.invoke.MethodHandles.publicLookup()
+					.findConstructor(mw == null || mw.IsStatic ? typeof(java.lang.AbstractMethodError) : typeof(java.lang.IllegalAccessError),
+						java.lang.invoke.MethodType.methodType(typeof(void), typeof(string)))
+					.bindTo(tw.Name + ".Invoke" + sig);
+				return Delegate.CreateDelegate(delegateType,
+						java.lang.invoke.MethodHandles.dropArguments(
+							java.lang.invoke.MethodHandles.foldArguments(java.lang.invoke.MethodHandles.throwException(methodType.returnType(), exception.type().returnType()), exception),
+							0, methodType.parameterArray()).vmtarget, "Invoke");
+#else
 				MethodInfo invoke = delegateType.GetMethod("Invoke");
 				ParameterInfo[] parameters = invoke.GetParameters();
 				Type[] parameterTypes = new Type[parameters.Length + 1];
@@ -449,109 +455,106 @@ namespace IKVM.Runtime
 				ilgen.Emit(System.Reflection.Emit.OpCodes.Throw);
 				ilgen.DoEmit();
 				return dm.CreateDelegate(delegateType, obj);
+#endif
 			}
 			else
 			{
 				mw.ResolveMethod();
 				return Delegate.CreateDelegate(delegateType, obj, (MethodInfo)mw.GetMethod());
 			}
-		}
-#else
-		[DebuggerStepThroughAttribute]
-		public static object DynamicCast(object obj, RuntimeTypeHandle type, string clazz)
-		{
-			return null;
-		}
-
-		[DebuggerStepThroughAttribute]
-		public static bool DynamicInstanceOf(object obj, RuntimeTypeHandle type, string clazz)
-		{
-			return false;
-		}
-
-		[DebuggerStepThroughAttribute]
-		public static Type DynamicGetTypeAsExceptionType(RuntimeTypeHandle type, string clazz)
-		{
-			return null;
+#endif
 		}
 
 		[DebuggerStepThrough]
-		public static Delegate DynamicCreateDelegate(object obj, Type delegateType)
+		public static ikvm.@internal.CallerID DynamicCallerID(object capability)
 		{
+			return ((DynamicCallerIDProvider)capability).GetCallerID();
+		}
+
+		[DebuggerStepThrough]
+		public static java.lang.invoke.MethodHandle DynamicEraseInvokeExact(java.lang.invoke.MethodHandle mh, java.lang.invoke.MethodType expected, java.lang.invoke.MethodType target)
+		{
+#if FIRST_PASS
 			return null;
+#else
+			if (mh.type() != expected)
+			{
+				throw new java.lang.invoke.WrongMethodTypeException();
+			}
+			return java.lang.invoke.MethodHandles.explicitCastArguments(mh, target);
+#endif
 		}
-#endif //!FIRST_PASS
 
 		[DebuggerStepThroughAttribute]
 		public static int f2i(float f)
 		{
-			if(f <= int.MinValue)
+			if (f > int.MinValue && f < int.MaxValue)
+			{
+				return (int)f;
+			}
+			if (f <= int.MinValue)
 			{
 				return int.MinValue;
 			}
-			if(f >= int.MaxValue)
+			if (f >= int.MaxValue)
 			{
 				return int.MaxValue;
 			}
-			if(float.IsNaN(f))
-			{
-				return 0;
-			}
-			return (int)f;
+			return 0;
 		}
 
 		[DebuggerStepThroughAttribute]
 		public static long f2l(float f)
 		{
-			if(f <= long.MinValue)
+			if (f > long.MinValue && f < long.MaxValue)
+			{
+				return (long)f;
+			}
+			if (f <= long.MinValue)
 			{
 				return long.MinValue;
 			}
-			if(f >= long.MaxValue)
+			if (f >= long.MaxValue)
 			{
 				return long.MaxValue;
 			}
-			if(float.IsNaN(f))
-			{
-				return 0;
-			}
-			return (long)f;
+			return 0;
 		}
 
 		[DebuggerStepThroughAttribute]
 		public static int d2i(double d)
 		{
-			if(d <= int.MinValue)
+			if (d > int.MinValue && d < int.MaxValue)
+			{
+				return (int)d;
+			}
+			if (d <= int.MinValue)
 			{
 				return int.MinValue;
 			}
-			if(d >= int.MaxValue)
+			if (d >= int.MaxValue)
 			{
 				return int.MaxValue;
 			}
-			if(double.IsNaN(d))
-			{
-				return 0;
-			}
-			return (int)d;
+			return 0;
 		}
 
 		[DebuggerStepThroughAttribute]
 		public static long d2l(double d)
 		{
-			if(d <= long.MinValue)
+			if (d > long.MinValue && d < long.MaxValue)
+			{
+				return (long)d;
+			}
+			if (d <= long.MinValue)
 			{
 				return long.MinValue;
 			}
-			if(d >= long.MaxValue)
+			if (d >= long.MaxValue)
 			{
 				return long.MaxValue;
 			}
-			if(double.IsNaN(d))
-			{
-				return 0;
-			}
-			return (long)d;
+			return 0;
 		}
 
 		// This is used by static JNI and synchronized methods that need a class object
@@ -801,22 +804,50 @@ namespace IKVM.Runtime
 
 		public static long VolatileRead(ref long v)
 		{
+#if NO_REF_EMIT && !FIRST_PASS
+			lock (VolatileLongDoubleFieldWrapper.lockObject)
+			{
+				return v;
+			}
+#else
 			return Interlocked.Read(ref v);
+#endif
 		}
 
 		public static void VolatileWrite(ref long v, long newValue)
 		{
+#if NO_REF_EMIT && !FIRST_PASS
+			lock (VolatileLongDoubleFieldWrapper.lockObject)
+			{
+				v = newValue;
+			}
+#else
 			Interlocked.Exchange(ref v, newValue);
+#endif
 		}
 
 		public static double VolatileRead(ref double v)
 		{
+#if NO_REF_EMIT && !FIRST_PASS
+			lock (VolatileLongDoubleFieldWrapper.lockObject)
+			{
+				return v;
+			}
+#else
 			return Interlocked.CompareExchange(ref v, 0.0, 0.0);
+#endif
 		}
 
 		public static void VolatileWrite(ref double v, double newValue)
 		{
+#if NO_REF_EMIT && !FIRST_PASS
+			lock (VolatileLongDoubleFieldWrapper.lockObject)
+			{
+				v = newValue;
+			}
+#else
 			Interlocked.Exchange(ref v, newValue);
+#endif
 		}
 
 		[System.Runtime.CompilerServices.MethodImpl(System.Runtime.CompilerServices.MethodImplOptions.NoInlining)]
@@ -868,32 +899,50 @@ namespace IKVM.Runtime
 			return ExceptionHelper.MapException<T>(x, (mode & MapFlags.NoRemapping) == 0, (mode & MapFlags.Unused) != 0);
 		}
 
+		[HideFromJava]
+		public static Exception DynamicMapException(Exception x, MapFlags mode, java.lang.Class exceptionClass)
+		{
+#if FIRST_PASS
+			return null;
+#else
+			TypeWrapper exceptionTypeWrapper = TypeWrapper.FromClass(exceptionClass);
+			mode &= ~MapFlags.NoRemapping;
+			if (exceptionTypeWrapper.IsSubTypeOf(CoreClasses.cli.System.Exception.Wrapper))
+			{
+				mode |= MapFlags.NoRemapping;
+			}
+			Type exceptionType = exceptionTypeWrapper == CoreClasses.java.lang.Throwable.Wrapper ? typeof(System.Exception) : exceptionTypeWrapper.TypeAsBaseType;
+			return (Exception)ByteCodeHelperMethods.mapException.MakeGenericMethod(exceptionType).Invoke(null, new object[] { x, mode });
+#endif
+		}
+
 		public static T GetDelegateForInvokeExact<T>(global::java.lang.invoke.MethodHandle h)
 			where T : class
 		{
 #if FIRST_PASS
 			return null;
 #else
-			T del = h.vmtarget as T;
+			T del = h._invokeExactDelegate as T;
 			if (del == null)
 			{
-				throw new global::java.lang.invoke.WrongMethodTypeException();
+				del = MethodHandleUtil.GetDelegateForInvokeExact<T>(h);
 			}
 			return del;
 #endif
 		}
 
-		public static T GetDelegateForInvoke<T>(global::java.lang.invoke.MethodHandle h, ref InvokeCache<T> cache)
+		public static T GetDelegateForInvoke<T>(global::java.lang.invoke.MethodHandle h, java.lang.invoke.MethodType realType, ref InvokeCache<T> cache)
 			where T : class
 		{
 #if FIRST_PASS
 			return null;
 #else
-			if (cache.type == h.type() && cache.del != null)
+			T del;
+			if (cache.Type == h.type() && (del = (h.isVarargsCollector() ? cache.varArg : cache.fixedArg)) != null)
 			{
-				return cache.del;
+				return del;
 			}
-			T del = h.vmtarget as T;
+			del = h.form.vmentry.vmtarget as T;
 			if (del == null)
 			{
 				global::java.lang.invoke.MethodHandle adapter = global::java.lang.invoke.MethodHandles.exactInvoker(h.type());
@@ -901,11 +950,39 @@ namespace IKVM.Runtime
 				{
 					adapter = adapter.asVarargsCollector(h.type().parameterType(h.type().parameterCount() - 1));
 				}
-				del = (T)adapter.asType(MethodHandleUtil.GetDelegateMethodType(typeof(T))).vmtarget;
-				if (Interlocked.CompareExchange(ref cache.type, h.type(), null) == null)
+				// if realType is set, the delegate contains erased unloadable types
+				if (realType != null)
 				{
-					cache.del = del;
+					adapter = adapter.asType(realType.insertParameterTypes(0, ikvm.@internal.ClassLiteral<java.lang.invoke.MethodHandle>.Value)).asFixedArity();
 				}
+				adapter = adapter.asType(MethodHandleUtil.GetDelegateMethodType(typeof(T)));
+				del = GetDelegateForInvokeExact<T>(adapter);
+				if (cache.TrySetType(h.type()))
+				{
+					if (h.isVarargsCollector())
+					{
+						cache.varArg = del;
+					}
+					else
+					{
+						cache.fixedArg = del;
+					}
+				}
+			}
+			return del;
+#endif
+		}
+
+		public static T GetDelegateForInvokeBasic<T>(java.lang.invoke.MethodHandle h)
+			where T : class
+		{
+#if FIRST_PASS
+			return null;
+#else
+			T del = h.form.vmentry.vmtarget as T;
+			if (del == null)
+			{
+				del = MethodHandleUtil.GetVoidAdapter(h.form.vmentry) as T;
 			}
 			return del;
 #endif
@@ -921,31 +998,24 @@ namespace IKVM.Runtime
 #endif
 		}
 
-#if !FIRST_PASS
-		sealed class ConstantMethodHandle : java.lang.invoke.MethodHandle
-		{
-			internal ConstantMethodHandle(Delegate del)
-				: base(MethodHandleUtil.GetDelegateMethodType(del.GetType()))
-			{
-				this.vmtarget = del;
-			}
-		}
-#endif
-
-		public static java.lang.invoke.MethodHandle MethodHandleFromDelegate(Delegate del)
-		{
-#if FIRST_PASS
-			return null;
-#else
-			return new ConstantMethodHandle(del);
-#endif
-		}
-
 		[HideFromJava]
 		public static void LinkIndyCallSite<T>(ref IndyCallSite<T> site, java.lang.invoke.CallSite cs, Exception x)
 			where T : class // Delegate
 		{
 #if !FIRST_PASS
+			// when a CallSite is first constructed, it doesn't call MethodHandleNatives.setCallSiteTargetNormal(),
+			// so we have to check if we need to initialize it here (i.e. attach an IndyCallSite<T> to it)
+			if (cs != null)
+			{
+				if (cs.ics == null)
+				{
+					Java_java_lang_invoke_MethodHandleNatives.InitializeCallSite(cs);
+				}
+				lock (cs.ics)
+				{
+					cs.ics.SetTarget(cs.target);
+				}
+			}
 			IndyCallSite<T> ics;
 			if (x != null || cs == null || (ics = cs.ics as IndyCallSite<T>) == null)
 			{
@@ -953,14 +1023,17 @@ namespace IKVM.Runtime
 					? (Exception)new java.lang.ClassCastException("bootstrap method failed to produce a CallSite")
 					: new java.lang.invoke.WrongMethodTypeException()), MapFlags.None);
 				java.lang.invoke.MethodType type = LoadMethodType<T>();
-				ics = new IndyCallSite<T>((T)
+				java.lang.invoke.MethodHandle exc = x is java.lang.BootstrapMethodError
+					? java.lang.invoke.MethodHandles.constant(typeof(java.lang.BootstrapMethodError), x)
+					: java.lang.invoke.MethodHandles.publicLookup().findConstructor(typeof(java.lang.BootstrapMethodError), java.lang.invoke.MethodType.methodType(typeof(void), typeof(string), typeof(Exception)))
+						.bindTo("call site initialization exception").bindTo(x);
+				ics = new IndyCallSite<T>();
+				((IIndyCallSite)ics).SetTarget(
 					java.lang.invoke.MethodHandles.dropArguments(
 						java.lang.invoke.MethodHandles.foldArguments(
 							java.lang.invoke.MethodHandles.throwException(type.returnType(), typeof(java.lang.BootstrapMethodError)),
-								new ConstantMethodHandle((MH<Exception, java.lang.BootstrapMethodError>)CreateBootstrapException).bindTo(x)),
-						0, type.parameterArray())
-					.vmtarget,
-					false);
+								exc),
+						0, type.parameterArray()));
 			}
 			IndyCallSite<T> curr = site;
 			if (curr.IsBootstrap)
@@ -970,23 +1043,65 @@ namespace IKVM.Runtime
 #endif
 		}
 
-#if !FIRST_PASS
 		[HideFromJava]
-		private static java.lang.BootstrapMethodError CreateBootstrapException(Exception x)
+		public static void DynamicLinkIndyCallSite<T>(ref IndyCallSite<T> site, java.lang.invoke.CallSite cs, Exception x, string signature, ikvm.@internal.CallerID callerID)
+			where T : class // Delegate
 		{
-			if (x is java.lang.BootstrapMethodError)
+#if !FIRST_PASS
+			// when a CallSite is first constructed, it doesn't call MethodHandleNatives.setCallSiteTargetNormal(),
+			// so we have to check if we need to initialize it here (i.e. attach an IndyCallSite<T> to it)
+			if (cs != null)
 			{
-				return (java.lang.BootstrapMethodError)x;
+				if (cs.ics == null)
+				{
+					Java_java_lang_invoke_MethodHandleNatives.InitializeCallSite(cs);
+				}
+				lock (cs.ics)
+				{
+					cs.ics.SetTarget(cs.target);
+				}
+			}
+			java.lang.invoke.MethodType typeCache = null;
+			IndyCallSite<T> ics;
+			if (x != null || cs == null || cs.type() != DynamicLoadMethodType(ref typeCache, signature, callerID))
+			{
+				x = MapException<Exception>(x ?? (cs == null
+					? (Exception)new java.lang.ClassCastException("bootstrap method failed to produce a CallSite")
+					: new java.lang.invoke.WrongMethodTypeException()), MapFlags.None);
+				java.lang.invoke.MethodType type = LoadMethodType<T>();
+				java.lang.invoke.MethodHandle exc = x is java.lang.BootstrapMethodError
+					? java.lang.invoke.MethodHandles.constant(typeof(java.lang.BootstrapMethodError), x)
+					: java.lang.invoke.MethodHandles.publicLookup().findConstructor(typeof(java.lang.BootstrapMethodError), java.lang.invoke.MethodType.methodType(typeof(void), typeof(string), typeof(Exception)))
+						.bindTo("call site initialization exception").bindTo(x);
+				ics = new IndyCallSite<T>();
+				((IIndyCallSite)ics).SetTarget(
+					java.lang.invoke.MethodHandles.dropArguments(
+						java.lang.invoke.MethodHandles.foldArguments(
+							java.lang.invoke.MethodHandles.throwException(type.returnType(), typeof(java.lang.BootstrapMethodError)),
+								exc),
+						0, type.parameterArray()));
+			}
+			else
+			{
+				ics = new IndyCallSite<T>();
+				((IIndyCallSite)ics).SetTarget(cs.dynamicInvoker().asType(LoadMethodType<T>()));
+			}
+			IndyCallSite<T> curr = site;
+			if (curr.IsBootstrap)
+			{
+				Interlocked.CompareExchange(ref site, ics, curr);
 			}
-			return new java.lang.BootstrapMethodError("call site initialization exception", x);
-		}
 #endif
+		}
+	}
+
+	interface IIndyCallSite
+	{
+		void SetTarget(java.lang.invoke.MethodHandle mh);
 	}
 
 	public sealed class IndyCallSite<T>
-#if !FIRST_PASS
-		: java.lang.invoke.CallSite.IndyCallSite
-#endif
+		: IIndyCallSite
 		where T : class // Delegate
 	{
 		internal readonly bool IsBootstrap;
@@ -1002,13 +1117,13 @@ namespace IKVM.Runtime
 			this.target = target;
 		}
 
-#if !FIRST_PASS
-		void java.lang.invoke.CallSite.IndyCallSite.setTarget(object target)
+		void IIndyCallSite.SetTarget(java.lang.invoke.MethodHandle mh)
 		{
-			this.target = (T)target;
-		}
+#if !FIRST_PASS
+			target = ByteCodeHelper.GetDelegateForInvokeExact<T>(mh);
 #endif
-
+		}
+		
 		public static IndyCallSite<T> CreateBootstrap(T bootstrap)
 		{
 			return new IndyCallSite<T>(bootstrap, true);
@@ -1023,8 +1138,38 @@ namespace IKVM.Runtime
 	public struct InvokeCache<T>
 		where T : class
 	{
-		internal global::java.lang.invoke.MethodType type;
-		internal T del;
+#if CLASSGC
+		private WeakReference weakRef;
+
+		internal java.lang.invoke.MethodType Type
+		{
+			get { return weakRef == null ? null : (java.lang.invoke.MethodType)weakRef.Target; }
+		}
+
+		internal bool TrySetType(java.lang.invoke.MethodType newType)
+		{
+			if (weakRef == null)
+			{
+				return Interlocked.CompareExchange(ref weakRef, new WeakReference(newType), null) == null;
+			}
+			return Type == newType;
+		}
+#else
+		private java.lang.invoke.MethodType type;
+
+		internal java.lang.invoke.MethodType Type
+		{
+			get { return type; }
+		}
+
+		internal bool TrySetType(java.lang.invoke.MethodType newType)
+		{
+			Interlocked.CompareExchange(ref type, newType, null);
+			return type == newType;
+		}
+#endif
+		internal T fixedArg;
+		internal T varArg;
 	}
 
 	[StructLayout(LayoutKind.Explicit)]
@@ -1112,4 +1257,9 @@ namespace IKVM.Runtime
 	public delegate TResult MH<T1, T2, T3, T4, T5, T6, TResult>(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5, T6 t6);
 	public delegate TResult MH<T1, T2, T3, T4, T5, T6, T7, TResult>(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5, T6 t6, T7 t7);
 	public delegate TResult MH<T1, T2, T3, T4, T5, T6, T7, T8, TResult>(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5, T6 t6, T7 t7, T8 t8);
+
+	public static class LiveObjectHolder<T>
+	{
+		public static object[] values;
+	}
 }
